"""
Classes that parse Label Studio exports.

Does not know about Chart Review concepts like config files.
Simply parses and exposes Label Studio concepts.

Some of this parsing is a little generous (like allowing missing id, origin, type, or from_name
fields), mostly to accommodate tests without fully detailed input.
(Which, I think is fine - generous parsing is good anyway, and tests with fully specified exports
might make it harder to understand what the test is focusing on.)
"""

import dataclasses

from chart_review import common, defines, errors


@dataclasses.dataclass(kw_only=True)
class Mention:
    """A piece of text and labels, very similar to LabeledText."""

    id: str
    text: str
    labels: defines.LabelSet
    from_name: str

    @staticmethod
    def parse(entry: dict) -> "Mention":
        # Check where we're going to find the labels/tags
        match entry.get("type", "labels").casefold():
            case "labels":
                # Looks like:
                # "value": {
                #   "text": "patient has infection",
                #   "labels": ["Infection"]
                # },
                field = "labels"
            case "choices":
                # Looks like:
                # "value": {
                #   "text": "patient is deceased",
                #   "choices": ["False"]
                # },
                field = "choices"
            case "textarea":
                # Looks like:
                # "value": {
                #   "text": ["free form"],
                # },
                field = "text"
            case _:
                raise ValueError(f"Unrecognized Label Studio result type '{entry.get('type')}'.")

        value = entry.get("value", {})
        text = value.get("text", "") if field != "text" else ""
        labels = set(defines.Label(x) for x in value.get(field, []))
        return Mention(
            id=entry.get("id", ""), text=text, labels=labels, from_name=entry.get("from_name", "")
        )


@dataclasses.dataclass(kw_only=True)
class Annotation:
    """All of a single source's mentions"""

    author: int
    mentions: list[Mention] = dataclasses.field(default_factory=list)

    @staticmethod
    def parse(entry: dict, data_keys: set[str]) -> "Annotation | None":
        author = entry.get("completed_by")
        if author is None:
            return None  # we don't know who annotated this!

        # Labels can be nested - e.g. there might be a toplevel label "Illness" and a sublabel
        # (maybe called "Illness Confirmed?") with a three-way choice of "confirmed", "suspected",
        # and "none of the above".
        # The way that would shows up as a mention in an export is a little odd (to my mind):
        # {
        #   "value": {
        #     "text": "text from note",
        #     "labels": ["Illness"]
        #   },
        #   "id": "HI1y_tlNwu",
        #   "from_name": "label",
        #   "to_name": "text",
        #   "type": "labels",
        # },
        # {
        #   "value": {
        #     "text": "text from note",
        #     "choices": ["confirmed"]
        #   },
        #   "id": "HI1y_tlNwu",
        #   "from_name": "Illness Confirmed?",
        #   "to_name": "text",
        #   "type": "choices",
        # },
        #
        # So you can see there that the "id" field is re-used, and the sublabel refers to the
        # parent via "from_name".

        # When parsing here, we'll first look for the toplevel entries (identifiable by a
        # "from_name" pointing at a key in data_keys). Then do a second pass for any sublabels and
        # adjust the parent with the extra info.
        mentions: list[Mention] = []
        toplevels: dict[str, Mention] = {}
        sublabels: list[Mention] = []
        for result in entry.get("result", []):
            mention = Mention.parse(result)
            if not mention.from_name or mention.from_name in data_keys:
                # This is a toplevel mention
                toplevels[mention.id] = mention
                mentions.append(mention)
            else:
                # It's a sublabel, set it aside for a second
                sublabels.append(mention)

        # Now match up the sublabels
        base_sets: dict[str, defines.LabelSet] = {}
        for sublabel in sublabels:
            if sublabel.id not in toplevels:
                raise ValueError(f"Unrecognized sublabel ID '{sublabel.id}'.")
            toplevel = toplevels[sublabel.id]

            # Wipe out any toplevel tags (existence of a sublabel implies no toplevel labels)
            if toplevel.id not in base_sets:
                base_sets[toplevel.id] = toplevel.labels
                toplevel.labels = set()

            # Now merge in new labels (preserving other fields like `text` from toplevel entry)
            base_labels = base_sets[toplevel.id]
            for label in sublabel.labels:
                for base_label in base_labels:
                    toplevel.labels.add(
                        defines.Label(base_label.label, sublabel.from_name, label.label)
                    )

        return Annotation(author=author, mentions=mentions)


@dataclasses.dataclass(kw_only=True)
class Note:
    """All of a single note's annotations"""

    note_id: int
    annotations: list[Annotation] = dataclasses.field(default_factory=list)

    # metadata
    docref_mappings: dict[str, str] = dataclasses.field(default_factory=dict)
    encounter_id: str | None = None
    anon_encounter_id: str | None = None

    @staticmethod
    def parse(entry: dict) -> "Note":
        metadata = entry.get("data", {})
        docref_mappings = metadata.get("docref_mappings", {})
        encounter_id = metadata.get("encounter_id") or metadata.get("enc_id")  # old name
        anon_encounter_id = metadata.get("anon_encounter_id") or metadata.get("anon_id")  # old name

        data_keys = set(metadata.keys())
        annotations = [
            Annotation.parse(x, data_keys=data_keys) for x in entry.get("annotations", [])
        ]
        annotations = list(filter(None, annotations))  # parse() returns None if we should skip

        return Note(
            note_id=entry["id"],
            annotations=annotations,
            docref_mappings=docref_mappings,
            encounter_id=encounter_id,
            anon_encounter_id=anon_encounter_id,
        )


class ExportFile:
    """Parse information from a single Label Studio export file."""

    def __init__(self, path: str):
        try:
            data = common.read_json(path)
        except Exception as exc:
            errors.exit_for_invalid_project(str(exc))

        self._notes = [Note.parse(x) for x in data]

    @property
    def notes(self) -> list[Note]:
        return self._notes



"""Tests for commands/labels.py"""

import tempfile

from chart_review import common
from tests import base


class TestLabels(base.TestCase):
    """Test case for the top-level labels code"""

    def test_labels(self):
        stdout = self.run_cli("labels", path=f"{self.DATA_DIR}/cold")

        self.assertEqual(
            """╭───────────┬──────────┬─────────────╮
│ Annotator │ Label    │ Chart Count │
├───────────┼──────────┼─────────────┤
│ Any       │ Cough    │ 2           │
│ Any       │ Fatigue  │ 3           │
│ Any       │ Headache │ 3           │
├───────────┼──────────┼─────────────┤
│ jane      │ Cough    │ 1           │
│ jane      │ Fatigue  │ 2           │
│ jane      │ Headache │ 2           │
├───────────┼──────────┼─────────────┤
│ jill      │ Cough    │ 2           │
│ jill      │ Fatigue  │ 3           │
│ jill      │ Headache │ 0           │
├───────────┼──────────┼─────────────┤
│ john      │ Cough    │ 1           │
│ john      │ Fatigue  │ 2           │
│ john      │ Headache │ 2           │
╰───────────┴──────────┴─────────────╯
""",
            stdout,
        )

    def test_labels_grouped(self):
        """Verify that we only show final grouped labels, not intermediate ones"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "labels": ["fever", "rash", "recent"],
                    "grouped-labels": {"symptoms": ["fever", "rash"]},
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [],
            )
            stdout = self.run_cli("labels", path=tmpdir)

        self.assertEqual(
            """╭───────────┬──────────┬─────────────╮
│ Annotator │ Label    │ Chart Count │
├───────────┼──────────┼─────────────┤
│ Any       │ recent   │ 0           │
│ Any       │ symptoms │ 0           │
╰───────────┴──────────┴─────────────╯
""",
            stdout,
        )

    def test_labels_ignored(self):
        """Verify that we show info on ignored notes"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "ignore": [3, 4, 6],
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {"id": 3},
                    {"id": 4},
                    {"id": 5},
                    {"id": 6},
                ],
            )
            stdout = self.run_cli("labels", path=tmpdir)

        self.assertEqual(
            "Ignoring 3 charts (3–4, 6)",
            stdout.splitlines()[-1].strip(),
        )

    def test_labels_csv(self):
        """Verify that can print in CSV format"""
        stdout = self.run_cli("labels", "--csv", path=f"{self.DATA_DIR}/external")

        self.assertEqual(
            [
                "annotator,label,chart_count",
                "Any,happy,1",
                "Any,sad,1",
                "human,happy,1",
                "human,sad,1",
                "icd10-doc,happy,1",
                "icd10-doc,sad,1",
                "icd10-enc,happy,1",
                "icd10-enc,sad,1",
            ],
            stdout.splitlines(),
        )

    def test_sublabels(self):
        output = self.run_cli("labels", path=f"{self.DATA_DIR}/sublabels")
        self.assertEqual(
            output,
            """╭───────────┬────────────────────────────────┬─────────────╮
│ Annotator │ Label                          │ Chart Count │
├───────────┼────────────────────────────────┼─────────────┤
│ Any       │ Deceased                       │ 1           │
│ Any       │ Deceased → False               │ 1           │
│ Any       │ Deceased → Datetime → 11/12/25 │ 1           │
│ Any       │ Deceased → Datetime → 11/13/25 │ 1           │
│ Any       │ Fungal → Confirmed             │ 1           │
│ Any       │ Infection                      │ 1           │
│ Any       │ Infection → Confirmed          │ 1           │
│ Any       │ Infection → Suspected          │ 1           │
├───────────┼────────────────────────────────┼─────────────┤
│ alice     │ Deceased                       │ 1           │
│ alice     │ Deceased → False               │ 1           │
│ alice     │ Deceased → Datetime → 11/12/25 │ 1           │
│ alice     │ Deceased → Datetime → 11/13/25 │ 0           │
│ alice     │ Fungal → Confirmed             │ 1           │
│ alice     │ Infection                      │ 0           │
│ alice     │ Infection → Confirmed          │ 0           │
│ alice     │ Infection → Suspected          │ 1           │
├───────────┼────────────────────────────────┼─────────────┤
│ bob       │ Deceased                       │ 0           │
│ bob       │ Deceased → False               │ 1           │
│ bob       │ Deceased → Datetime → 11/12/25 │ 0           │
│ bob       │ Deceased → Datetime → 11/13/25 │ 1           │
│ bob       │ Fungal → Confirmed             │ 1           │
│ bob       │ Infection                      │ 1           │
│ bob       │ Infection → Confirmed          │ 0           │
│ bob       │ Infection → Suspected          │ 1           │
├───────────┼────────────────────────────────┼─────────────┤
│ carla     │ Deceased                       │ 0           │
│ carla     │ Deceased → False               │ 1           │
│ carla     │ Deceased → Datetime → 11/12/25 │ 1           │
│ carla     │ Deceased → Datetime → 11/13/25 │ 0           │
│ carla     │ Fungal → Confirmed             │ 1           │
│ carla     │ Infection                      │ 0           │
│ carla     │ Infection → Confirmed          │ 1           │
│ carla     │ Infection → Suspected          │ 0           │
╰───────────┴────────────────────────────────┴─────────────╯
""",
        )



"""Methods for McNemar calculations."""

import argparse

import rich
import rich.box
import rich.table
import rich.text
from scipy.stats import binom, chi2

from chart_review import cli_utils, console_utils


def make_subparser(parser: argparse.ArgumentParser) -> None:
    cli_utils.add_project_args(parser)
    cli_utils.add_output_args(parser)
    parser.add_argument("truth_annotator")
    parser.add_argument("annotator1")
    parser.add_argument("annotator2")
    parser.set_defaults(func=print_mcnemar)


def print_mcnemar(args: argparse.Namespace) -> None:
    """
    McNemar calculation between three annotators (one truth, two normal).

    More information:
     https://en.wikipedia.org/wiki/McNemar's_test
     https://pmc.ncbi.nlm.nih.gov/articles/PMC3716987/
    """
    reader = cli_utils.get_cohort_reader(args)
    truth = args.truth_annotator
    annotator1 = args.annotator1
    annotator2 = args.annotator2

    all_people = {truth, annotator1, annotator2}
    annotators = [annotator1, annotator2]

    for annotator in all_people:
        if annotator not in reader.note_range:
            raise ValueError(f"Unrecognized annotator '{annotator}'")

    if len(all_people) != 3:
        raise ValueError("Can’t compare the same annotator with themselves.")

    # Grab the intersection of ranges
    note_range = reader.note_range[truth]
    for annotator in annotators:
        note_range &= reader.note_range[annotator]

    labels = [None, *sorted(reader.class_labels)]

    # Calculate contingency tables
    matrices = {
        label: reader.contingency_table(truth, annotator1, annotator2, note_range, label)
        for label in labels
    }

    console = rich.get_console()

    table = cli_utils.create_table(
        "McNemar", "P-value", "BC", "OL", "OR", "BW", "Label", dense=True
    )
    empty_val = "" if args.csv else "N/A"
    for label in labels:
        m = matrices[label]
        mcn, pval = _mcnemar(len(m["OL"]), len(m["OR"]), continuity_correction=True)
        table.add_row(
            empty_val if mcn is None else _small_float(mcn),
            _small_float(pval),
            str(len(m["BC"])),
            str(len(m["OL"])),
            str(len(m["OR"])),
            str(len(m["BW"])),
            str(label) if label else "*",
        )

    if args.csv:
        cli_utils.print_table_as_csv(table)
        return

    # OK we aren't printing a CSV file to stdout, so we can include a bit more explanation
    # as a little header to the real results.
    note_count = len(note_range)
    chart_word = "chart" if note_count == 1 else "charts"
    pretty_ranges = f" ({console_utils.pretty_note_range(note_range)})" if note_count > 0 else ""
    console.print(f"Comparing {note_count} {chart_word}{pretty_ranges}")
    console.print(f"Truth: {truth}")
    console.print(f"Annotators: {', '.join(annotators)}")
    console.print()

    console.print(table)


def _small_float(number: float) -> str:
    rounded = str(round(number, 3))
    if number == 0 or rounded != "0.0":
        return rounded

    # Too small for simple rounding, use scientific notation instead
    return f"{number:.2e}"


# From https://en.wikipedia.org/wiki/McNemar's_test
# Licensed as CC-BY-SA-4.0
# With small tweaks to fit our style and return the McNemar test value.
def _mcnemar(b: int, c: int, continuity_correction: bool = False) -> tuple[float | None, float]:
    n_min, n_max = sorted([b, c])
    corr = 1 if continuity_correction else 0
    if (n_min + n_max) < 25:
        chi2_statistic = None
        pvalue = 2 * binom.cdf(n_min, n_min + n_max, 0.5) - binom.pmf(n_min, n_min + n_max, 0.5)
    else:
        chi2_statistic = (abs(n_min - n_max) - corr) ** 2 / (n_min + n_max)
        pvalue = chi2.sf(chi2_statistic, 1)
    return chi2_statistic, pvalue



import argparse

import rich
import rich.box
import rich.table
import rich.text

from chart_review import cli_utils, console_utils


def make_subparser(parser: argparse.ArgumentParser) -> None:
    cli_utils.add_project_args(parser)
    cli_utils.add_output_args(parser)
    parser.set_defaults(func=print_mentions)


def print_mentions(args: argparse.Namespace) -> None:
    """
    Print Label Studio export's mentions (text associated with the label).
    """
    reader = cli_utils.get_cohort_reader(args)

    table = cli_utils.create_table("Annotator", "Chart ID", "Mention", "Label")

    for annotator in sorted(reader.annotations.original_text_mentions, key=str.casefold):
        table.add_section()
        annotator_mentions = reader.annotations.original_text_mentions[annotator]
        for note_id, labeled_texts in annotator_mentions.items():
            # Gather all combos of text/label (i.e. all mentions) in this note
            note_mentions = set()
            for labeled_text in labeled_texts:
                text = labeled_text.text and labeled_text.text.casefold()
                for label in labeled_text.labels:
                    if label in reader.annotations.labels:
                        note_mentions.add((text, label))

            # Now add each mention to the table
            for note_mention in sorted(note_mentions, key=lambda m: (m[0], m[1])):
                table.add_row(annotator, str(note_id), note_mention[0], str(note_mention[1]))

    if args.csv:
        cli_utils.print_table_as_csv(table)
    else:
        rich.get_console().print(table)
        console_utils.print_ignored_charts(reader)



"""Utility methods"""

import json
import logging

###############################################################################
# Helper Functions: read/write JSON and text
###############################################################################


def read_json(path: str) -> dict | list[dict]:
    """
    Reads json from a file
    :param path: filesystem path
    :return: message: coded message
    """
    logging.debug("read_json() %s", path)

    with open(path) as f:
        return json.load(f, strict=False)


def write_json(path: str, data: dict | list, indent: int | None = 4) -> None:
    """
    Writes data to the given path, in json format
    :param path: filesystem path
    :param data: the structure to write to disk
    :param indent: whether and how much to indent the output
    """
    logging.debug("write_json() %s", path)

    with open(path, "w") as f:
        json.dump(data, f, indent=indent)


def read_text(path: str) -> str:
    """
    Reads data from the given path, in text format
    :param path: (currently filesystem path)
    :return: message: coded message
    """
    logging.debug("read_text() %s", path)

    with open(path) as f:
        return f.read()


def write_text(path: str, text: str) -> None:
    """
    Writes data to the given path, in text format
    :param path: filesystem path
    :param text: the text to write to disk
    """
    logging.debug("write_text() %s", path)

    with open(path, "w") as f:
        f.write(text)



"""Tests for config.py"""

import os
import tempfile

import ddt

from chart_review import common, config
from tests import base


@ddt.ddt
class TestProjectConfig(base.TestCase):
    """Test case for basic config parsing"""

    def make_config(self, conf_text: str, filename: str = "config.yaml") -> config.ProjectConfig:
        tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(tmpdir.cleanup)
        common.write_text(os.path.join(tmpdir.name, filename), conf_text)
        return config.ProjectConfig(tmpdir.name)

    @ddt.data(
        (
            "yaml",
            """
            labels:
                - cough | cough | sever
                - fever
            annotators:
                jane: 1
                john: 2
            ranges:
                jane: [3]
                john: [1, 3, 5]
        """,
        ),
        (
            "json",
            """
            {
                "labels": ["cough | cough | sever", "fever"],
                "annotators": {"jane": 1, "john": 2},
                "ranges": {"jane": [3], "john": [1, 3, 5]}
            }
        """,
        ),
    )
    @ddt.unpack
    def test_multiple_formats(self, suffix, text):
        """Verify that we can operate on multiple formats (like json & yaml)."""
        proj_config = self.make_config(text, filename=f"config.{suffix}")

        self.assertEqual(base.labels({"cough|cough|sever", "fever"}), proj_config.class_labels)
        self.assertEqual({1: "jane", 2: "john"}, proj_config.annotators)
        self.assertEqual({"jane": [3], "john": [1, 3, 5]}, proj_config.note_ranges)

    def test_range_syntax(self):
        """Verify that we support interesting note range syntax options."""
        with self.capture_stderr() as stderr:
            proj_config = self.make_config(
                """
                ranges:
                    bare_num: 1
                    string_num: "2"
                    range: 3-5
                    reference: bare_num
                    array: [1, "2", range]
                    invalid: [not_defined, 1]
            """
            )

        self.assertEqual(
            {
                "bare_num": [1],
                "string_num": [2],
                "range": [3, 4, 5],
                "reference": [1],
                "array": [1, 2, 3, 4, 5],
                "invalid": [1],
            },
            proj_config.note_ranges,
        )
        self.assertEqual(stderr.getvalue(), "Unknown note range 'not_defined'\n")

    def test_grouped_labels(self):
        """Verify that we grab the label config correctly."""
        proj_config = self.make_config(
            """
            grouped-labels:
                A|AX|AY: B|BX|BY
                C: [D, E]
            """
        )

        self.assertEqual(
            {
                base.Label("A", "AX", "AY"): base.LabelMatcher("B|BX|BY"),
                base.Label("C"): base.LabelMatcher("D", "E"),
            },
            proj_config.grouped_labels,
        )

    def test_implied_labels(self):
        """Verify that we grab the label config correctly."""
        proj_config = self.make_config(
            """
            implied-labels:
                A|AX|AY: B|BX|BY
                C: [D, E]
            """
        )

        self.assertEqual(
            {
                base.LabelMatcher("A|AX|AY"): {base.Label("B", "BX", "BY")},
                base.LabelMatcher("C"): base.labels({"D", "E"}),
            },
            proj_config.implied_labels,
        )

    def test_incomplete_label(self):
        with self.assertRaisesRegex(ValueError, "Sublabel name but no sublabel value provided"):
            proj_config = self.make_config("labels: [A|B]")
            print(proj_config.class_labels)  # reference the field so the labels get computed

    def test_label_with_extra_pipes(self):
        """They get put with the value side of things"""
        proj_config = self.make_config("labels: [A|B|C|D]")
        self.assertEqual(proj_config.class_labels, {base.Label("A", "B", "C|D")})

    def test_label_with_pipes_in_key_part(self):
        """They are not allowed"""
        # No normal way to make these, so just confirm we check for this manually
        with self.assertRaisesRegex(ValueError, "Invalid character found in label name: '|'."):
            base.Label("A|B")
        with self.assertRaisesRegex(ValueError, "Invalid character found in label name: '|'."):
            base.Label("A", "B|C")



"""Tests for commands/accuracy.py"""

import shutil
import tempfile

from chart_review import common
from tests import base


class TestAccuracy(base.TestCase):
    """Test case for the top-level accuracy code"""

    def test_default_output(self):
        stdout = self.run_cli("accuracy", "jill", "jane", path=f"{self.DATA_DIR}/cold")

        self.assertEqual(
            """Comparing 3 charts (1, 3–4)
Truth: jill
Annotator: jane
Macro F1: 0.556

F1     Sens  Spec  PPV  NPV   Kappa  TP  FN  TN  FP  Label   
0.667  0.75  0.6   0.6  0.75  0.341  3   1   3   2   *       
0.667  0.5   1.0   1.0  0.5   0.4    1   1   1   0   Cough   
1.0    1.0   1.0   1.0  1.0   1.0    2   0   1   0   Fatigue 
0      0     0     0    0     0      0   0   1   2   Headache
""",
            stdout,
        )

    def test_csv(self):
        stdout = self.run_cli("accuracy", "--csv", "jill", "jane", path=f"{self.DATA_DIR}/cold")

        self.assertEqual(
            [
                "f1,sens,spec,ppv,npv,kappa,tp,fn,tn,fp,label",
                "0.667,0.75,0.6,0.6,0.75,0.341,3,1,3,2,*",
                "0.667,0.5,1.0,1.0,0.5,0.4,1,1,1,0,Cough",
                "1.0,1.0,1.0,1.0,1.0,1.0,2,0,1,0,Fatigue",
                "0,0,0,0,0,0,0,0,1,2,Headache",
            ],
            stdout.splitlines(),
        )

    def test_save(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.copytree(f"{self.DATA_DIR}/cold", tmpdir, dirs_exist_ok=True)
            self.run_cli("accuracy", "--save", "jill", "jane", path=tmpdir)

            accuracy_json = common.read_json(f"{tmpdir}/accuracy-jill-jane.json")
            self.assertEqual(
                {
                    "F1": 0.667,
                    "Sens": 0.75,
                    "Spec": 0.6,
                    "PPV": 0.6,
                    "NPV": 0.75,
                    "Kappa": 0.341,
                    "TP": 3,
                    "FN": 1,
                    "TN": 3,
                    "FP": 2,
                    "Cough": {
                        "F1": 0.667,
                        "FN": 1,
                        "FP": 0,
                        "NPV": 0.5,
                        "PPV": 1.0,
                        "Sens": 0.5,
                        "Spec": 1.0,
                        "TN": 1,
                        "TP": 1,
                        "Kappa": 0.4,
                    },
                    "Fatigue": {
                        "F1": 1.0,
                        "FN": 0,
                        "FP": 0,
                        "NPV": 1.0,
                        "PPV": 1.0,
                        "Sens": 1.0,
                        "Spec": 1.0,
                        "TN": 1,
                        "TP": 2,
                        "Kappa": 1.0,
                    },
                    "Headache": {
                        "F1": 0,
                        "FN": 0,
                        "FP": 2,
                        "NPV": 0,
                        "PPV": 0,
                        "Sens": 0,
                        "Spec": 0,
                        "TN": 1,
                        "TP": 0,
                        "Kappa": 0,
                    },
                },
                accuracy_json,
            )

            accuracy_csv = common.read_text(f"{tmpdir}/accuracy-jill-jane.csv")
            self.assertEqual(
                """F1	Sens	Spec	PPV	NPV	Kappa	TP	FN	TN	FP	Label
0.667	0.75	0.6	0.6	0.75	0.341	3	1	3	2	*
0.667	0.5	1.0	1.0	0.5	0.4	1	1	1	0	Cough
1.0	1.0	1.0	1.0	1.0	1.0	2	0	1	0	Fatigue
0	0	0	0	0	0	0	0	1	2	Headache
""",
                accuracy_csv,
            )

    def test_save_and_csv_conflict(self):
        """Verify that --save and --csv can't run together"""
        with self.assertRaises(SystemExit) as cm:
            self.run_cli(
                "accuracy", "--save", "--csv", "jill", "jane", path=f"{self.DATA_DIR}/cold"
            )
        self.assertEqual(2, cm.exception.code)

    def test_verbose(self):
        output = self.run_cli("accuracy", "--verbose", "jill", "jane", path=f"{self.DATA_DIR}/cold")
        self.assertEqual(
            """Comparing 3 charts (1, 3–4)
Truth: jill
Annotator: jane

╭──────────┬──────────┬────────────────╮
│ Chart ID │ Label    │ Classification │
├──────────┼──────────┼────────────────┤
│ 1        │ Cough    │ TP             │
│ 1        │ Fatigue  │ TP             │
│ 1        │ Headache │ FP             │
├──────────┼──────────┼────────────────┤
│ 3        │ Cough    │ TN             │
│ 3        │ Fatigue  │ TN             │
│ 3        │ Headache │ TN             │
├──────────┼──────────┼────────────────┤
│ 4        │ Cough    │ FN             │
│ 4        │ Fatigue  │ TP             │
│ 4        │ Headache │ FP             │
╰──────────┴──────────┴────────────────╯
""",
            output,
        )

    def test_verbose_cvs(self):
        """Verify we can also print verbose results in CSV format"""
        stdout = self.run_cli(
            "accuracy", "--verbose", "--csv", "jill", "jane", path=f"{self.DATA_DIR}/cold"
        )
        self.assertEqual(
            [
                "chart_id,label,classification",
                "1,Cough,TP",
                "1,Fatigue,TP",
                "1,Headache,FP",
                "3,Cough,TN",
                "3,Fatigue,TN",
                "3,Headache,TN",
                "4,Cough,FN",
                "4,Fatigue,TP",
                "4,Headache,FP",
            ],
            stdout.splitlines(),
        )

    def test_bad_truth(self):
        """Verify that we do something suitable for bad arguments"""
        # Truth
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("accuracy", "nope1", "nope2", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Unrecognized annotator 'nope1'\n", stderr.getvalue())

        # Annotator
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("accuracy", "jill", "nope2", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Unrecognized annotator 'nope2'\n", stderr.getvalue())

    def test_same_annotators(self):
        """Verify that we do something suitable for bad arguments"""
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("accuracy", "jill", "jill", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Can’t compare the same annotator with themselves.\n", stderr.getvalue())

    def test_sublabels(self):
        output = self.run_cli("accuracy", "alice", "bob", path=f"{self.DATA_DIR}/sublabels")
        self.assertEqual(
            output,
            """Comparing 2 charts (1308–1309)
Truth: alice
Annotator: bob
Macro F1: 0.375

F1   Sens  Spec   PPV  NPV    Kappa  TP  FN  TN  FP  Label                      
0.6  0.6   0.778  0.6  0.778  0.378  3   2   7   2   *                          
0    0     0      0    0      0      0   1   1   0   Deceased                   
1.0  1.0   1.0    1.0  1.0    1.0    1   0   1   0   Deceased → False           
0    0     0      0    0      0      0   1   1   0   Deceased → Datetime →      
                                                     11/12/25                   
0    0     0      0    0      0      0   0   1   1   Deceased → Datetime →      
                                                     11/13/25                   
1.0  1.0   1.0    1.0  1.0    1.0    1   0   1   0   Fungal → Confirmed         
0    0     0      0    0      0      0   0   1   1   Infection                  
0    0     0      0    0      0      0   0   0   0   Infection → Confirmed      
1.0  1.0   1.0    1.0  1.0    1.0    1   0   1   0   Infection → Suspected      
""",
        )



"""Tests for commands/ids.py"""

import tempfile

from chart_review import common
from tests import base


class TestIDs(base.TestCase):
    """Test case for the top-level ids code"""

    def test_ids(self):
        """Verify that we pull IDs from all the places"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "annotators": {"carl": 1},
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {
                                        "value": {
                                            "labels": [
                                                "My Label",
                                            ],
                                        },
                                    },
                                ],
                            },
                        ],
                        # only orig encounter, no docref mappings
                        "data": {
                            "enc_id": "Orig",
                        },
                    },
                    {
                        "id": 2,
                        # only anon encounter, a single docref mapping
                        "data": {
                            "anon_id": "Anon",
                            "docref_mappings": {
                                "Orig": "Anon",
                            },
                        },
                    },
                    {
                        "id": 3,
                        # no encounter info, multiple docref mappings
                        "data": {
                            "docref_mappings": {
                                "Orig1": "Anon1",
                                "Orig2": "Anon2",
                            },
                        },
                    },
                    {
                        "id": 4,
                        # full encounter info, no docref mappings
                        # (using modern, full field names)
                        "data": {
                            "encounter_id": "a",
                            "anon_encounter_id": "b",
                        },
                    },
                    {
                        "id": 5,
                        # no metadata at all
                    },
                ],
            )
            stdout = self.run_cli("ids", path=tmpdir)

        self.assertEqual(
            [
                "╭──────────┬─────────────────────────┬─────────────────────────╮",
                "│ Chart ID │ Original FHIR ID        │ Anonymized FHIR ID      │",
                "├──────────┼─────────────────────────┼─────────────────────────┤",
                "│ 1        │ Encounter/Orig          │                         │",
                "│ 2        │                         │ Encounter/Anon          │",
                "│ 2        │ DocumentReference/Orig  │ DocumentReference/Anon  │",
                "│ 3        │ DocumentReference/Orig1 │ DocumentReference/Anon1 │",
                "│ 3        │ DocumentReference/Orig2 │ DocumentReference/Anon2 │",
                "│ 4        │ Encounter/a             │ Encounter/b             │",
                "│ 5        │                         │                         │",
                "╰──────────┴─────────────────────────┴─────────────────────────╯",
            ],
            stdout.splitlines(),
        )

    def test_ids_csv(self):
        """Verify that we can print CSV output"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(f"{tmpdir}/config.json", {})
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "data": {
                            # Verify that we quote correctly
                            "enc_id": "Orig,\\ 'Enc",
                            "anon_id": 'Anon "Enc',
                        },
                    },
                ],
            )
            stdout = self.run_cli("ids", "--csv", path=tmpdir)

        lines = stdout.splitlines()
        self.assertEqual(2, len(lines))
        self.assertEqual("chart_id,original_fhir_id,anonymized_fhir_id", lines[0])
        self.assertEqual('1,"Encounter/Orig,\\ \'Enc","Encounter/Anon ""Enc"', lines[1])



"""Run chart-review from the command-line"""

import argparse
import sys

from chart_review.commands import accuracy, default, frequency, ids, labels, mcnemar, mentions


def define_parser() -> argparse.ArgumentParser:
    """Fills out an argument parser with all the CLI options."""
    parser = argparse.ArgumentParser()
    default.make_subparser(parser)

    subparsers = parser.add_subparsers()
    accuracy.make_subparser(subparsers.add_parser("accuracy", help="calculate F1 and Kappa scores"))
    frequency.make_subparser(
        subparsers.add_parser("frequency", help="show counts of each text mention")
    )
    ids.make_subparser(subparsers.add_parser("ids", help="map Label Studio IDs to FHIR IDs"))
    labels.make_subparser(subparsers.add_parser("labels", help="show label usage by annotator"))
    mcnemar.make_subparser(subparsers.add_parser("mcnemar", help="calculate McNemar’s test"))
    mentions.make_subparser(subparsers.add_parser("mentions", help="show each mention of a label"))

    return parser


def main_cli(argv: list[str] | None = None) -> None:
    """Main entrypoint that wraps all the core program logic"""
    try:
        parser = define_parser()
        args = parser.parse_args(argv)
        args.func(args)
    except Exception as exc:
        print(str(exc), file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main_cli()  # pragma: no cover



from collections.abc import Iterable

from chart_review import agree, config, defines, external, simplify, studio


class CohortReader:
    """
    CohortReader converts a Label Studio export and a project config into a standard form.

    It also exposes some statistical helper methods.
    """

    def __init__(self, proj_config: config.ProjectConfig):
        """
        :param proj_config: parsed project configuration
        """
        self.config = proj_config
        self.project_dir = self.config.project_dir

        # Load exported annotations
        self.ls_export = studio.ExportFile(self.config.path("labelstudio-export.json"))

        self.annotations = simplify.simplify_export(self.ls_export, self.config)

        # Add a placeholder for any annotators that don't have mentions for some reason
        for annotator in self.config.annotators.values():
            self.annotations.mentions.setdefault(annotator, defines.Mentions())

        # Load external annotations (i.e. from NLP tags or ICD10 codes)
        for name, value in self.config.external_annotations.items():
            external.merge_external(self.annotations, self.ls_export, self.project_dir, name, value)

        # Consolidate/expand mentions based on config
        simplify.simplify_mentions(
            self.annotations,
            implied_labels=self.config.implied_labels,
            grouped_labels=self.config.grouped_labels,
        )

        # Calculate the final set of note ranges for each annotator
        self.note_range, self.ignored_notes = self._collect_note_ranges(self.ls_export)

        # Remove any ignored notes from the annotations, for ease of consuming code
        for note in self.ignored_notes:
            self.annotations.remove(note)

    def _collect_note_ranges(
        self, export: studio.ExportFile
    ) -> tuple[dict[str, defines.NoteSet], defines.NoteSet]:
        # Detect note ranges if they were not defined in the project config
        # (i.e. default to the full set of annotated notes)
        note_ranges = {k: set(v) for k, v in self.config.note_ranges.items()}
        for annotator, annotator_mentions in self.annotations.mentions.items():
            if annotator not in note_ranges:
                note_ranges[annotator] = set(annotator_mentions.keys())

        all_ls_notes = {note.note_id for note in export.notes}

        # Parse ignored IDs (might be note IDs, might be external IDs)
        ignored_notes = defines.NoteSet()
        for ignore_id in self.config.ignore:
            ls_id = external.external_id_to_label_studio_id(export, str(ignore_id))
            if ls_id is None:
                if isinstance(ignore_id, int):
                    ls_id = ignore_id  # must be direct note ID
                else:
                    # Must just be over-zealous excluding (like automatically from SQL)
                    continue  # pragma: no cover
            if ls_id in all_ls_notes:
                ignored_notes.add(ls_id)

        # Remove any invalid (ignored, non-existent) notes from the range sets
        for note_ids in note_ranges.values():
            note_ids.difference_update(ignored_notes)
            note_ids.intersection_update(all_ls_notes)

        return note_ranges, ignored_notes

    @property
    def class_labels(self) -> defines.LabelSet:
        return self.annotations.labels

    def _select_labels(self, label_pick: str | None = None) -> Iterable[str]:
        if label_pick:
            return [label_pick]
        else:
            return self.class_labels

    def confusion_matrix(
        self,
        truth: str,
        annotator: str,
        note_range: defines.NoteSet,
        label_pick: str | None = None,
    ) -> dict:
        """
        This is the rollup of counting each symptom only once, not multiple times.

        :param truth: annotator to use as the ground truth
        :param annotator: another annotator to compare with truth
        :param note_range: collection of LabelStudio document ID
        :param label_pick: (optional) of the CLASS_LABEL to score separately
        :return: dict
        """
        labels = self._select_labels(label_pick)
        return agree.confusion_matrix(
            self.annotations,
            truth,
            annotator,
            note_range,
            labels=labels,
        )

    def contingency_table(
        self,
        truth: str,
        annotator1: str,
        annotator2: str,
        note_range: defines.NoteSet,
        label_pick: str | None = None,
    ) -> dict:
        """
        Calculate a contingency table for truth and two annotators.

        https://en.wikipedia.org/wiki/Contingency_table

        :param truth: annotator to use as the ground truth
        :param annotator1: one annotator to compare
        :param annotator2: another annotator to compare
        :param note_range: collection of LabelStudio document ID
        :param label_pick: (optional) of the CLASS_LABEL to score separately
        :return: dict
        """
        labels = self._select_labels(label_pick)
        return agree.contingency_table(
            self.annotations,
            truth,
            annotator1,
            annotator2,
            note_range,
            labels=labels,
        )



"""Tests for external.py"""

import tempfile

from chart_review import cohort, common, config
from tests import base


class TestExternal(base.TestCase):
    """Test case for basic external ID merging"""

    def test_basic_read(self):
        reader = cohort.CohortReader(config.ProjectConfig(f"{self.DATA_DIR}/external"))

        self.assertEqual(
            {
                "human": {
                    1: {base.Label("happy"), base.Label("sad")},
                    # This was a note that didn't appear in the icd10 external annotations
                    # (and also didn't have a positive label by the human reviewer).
                    # Just here to test that it didn't screw anything up.
                    2: set(),
                    # This was an external annotation that said "saw it,
                    # but no labels for this note"
                    3: set(),
                },
                "icd10-doc": {1: base.labels({"happy", "hungry", "sad", "tired"}), 3: set()},
                "icd10-enc": {1: base.labels({"happy", "hungry", "sad", "tired"}), 3: set()},
            },
            reader.annotations.mentions,
        )

        # Confirm ranges got auto-detected for both human and icd10
        self.assertEqual(
            {
                "human": {1, 2, 3},
                "icd10-doc": {1, 3},
                "icd10-enc": {1, 3},
            },
            reader.note_range,
        )

    def test_assumes_encounter_id(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"ext": {"filename": "ext.csv"}}}
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {"id": 1, "data": {"enc_id": "could-be-either", "docref_mappings": {"a": "b"}}},
                    {"id": 2, "data": {"docref_mappings": {"could-be-either": "anon"}}},
                ],
            )
            common.write_text(f"{tmpdir}/ext.csv", "blarg,label\ncould-be-either,Confusing")
            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))

        self.assertEqual({"ext": {1}}, reader.note_range)

    def test_bad_annotator_config(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {"annotators": {"ext": "ext.csv"}},  # no filename:
            )
            common.write_json(f"{tmpdir}/labelstudio-export.json", [{"id": 1}])

            with self.assertRaisesRegex(ValueError, "^Did not understand config"):
                cohort.CohortReader(config.ProjectConfig(tmpdir))

    def test_missing_label_studio_data(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"ext": {"filename": "ext.csv"}}}
            )
            common.write_json(f"{tmpdir}/labelstudio-export.json", [{"id": 1, "data": {}}])
            common.write_text(f"{tmpdir}/ext.csv", "blarg,label\nenc-id,Fever")

            with self.assertRaisesRegex(ValueError, "^Your Label Studio export does not include"):
                cohort.CohortReader(config.ProjectConfig(tmpdir))

    def test_sublabels_happy_path(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"ext": {"filename": "ext.csv"}}}
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "data": {
                            "encounter_id": "enc",
                            "docref_mappings": {"DocumentReference/a": "DocumentReference/a_anon"},
                        },
                    },
                    {
                        "id": 2,
                        "data": {
                            "encounter_id": "enc",
                            "docref_mappings": {"DocumentReference/b": "DocumentReference/b_anon"},
                        },
                    },
                ],
            )
            common.write_text(
                f"{tmpdir}/ext.csv",
                """blarg,label,sublabel_name,sublabel_value,note_ref
                bogus,A,B,C,a
                bogus,,B,C,b""",  # empty 'label' col should give empty label set result
            )

            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))

        self.assertEqual(
            reader.annotations.mentions,
            {"ext": {1: base.labels({"A|B|C"}), 2: set()}},
        )

    def test_sublabels_no_id_col(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"ext": {"filename": "ext.csv"}}}
            )
            common.write_json(f"{tmpdir}/labelstudio-export.json", [{"id": 1, "data": {}}])
            common.write_text(
                f"{tmpdir}/ext.csv",
                """blarg,label,sublabel_name,sublabel_value
                bogus,A,B,C""",
            )

            with self.assertRaisesRegex(ValueError, "no resource ID column found"):
                cohort.CohortReader(config.ProjectConfig(tmpdir))

    def test_sublabels_no_label_col(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"ext": {"filename": "ext.csv"}}}
            )
            common.write_json(f"{tmpdir}/labelstudio-export.json", [{"id": 1, "data": {}}])
            common.write_text(
                f"{tmpdir}/ext.csv",
                """blarg,diagnosticreport_id,sublabel_name,sublabel_value
                bogus,id,B,C""",
            )

            with self.assertRaisesRegex(ValueError, "no 'label' column found"):
                cohort.CohortReader(config.ProjectConfig(tmpdir))

    def test_sublabels_no_sublabel_name_col(self):
        """This one works, you just can't specify sublabels, naturally"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"ext": {"filename": "ext.csv"}}}
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "data": {
                            "encounter_id": "enc",
                            "docref_mappings": {"DiagnosticReport/a": "DiagnosticReport/a_anon"},
                        },
                    }
                ],
            )
            common.write_text(
                f"{tmpdir}/ext.csv",
                """blarg,diagnosticreport_id,label
                bogus,a,Label1""",
            )

            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))

        self.assertEqual(reader.annotations.mentions, {"ext": {1: base.labels({"Label1"})}})

    def test_sublabels_no_sublabel_value_col(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"ext": {"filename": "ext.csv"}}}
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "data": {
                            "encounter_id": "enc",
                            "docref_mappings": {"DiagnosticReport/a": "DiagnosticReport/a_anon"},
                        },
                    }
                ],
            )
            common.write_text(
                f"{tmpdir}/ext.csv",
                """blarg,diagnosticreport_id,label,sublabel_name
                bogus,a,Label1,Sub1""",
            )

            with self.assertRaisesRegex(ValueError, "no 'sublabel_value' column found"):
                cohort.CohortReader(config.ProjectConfig(tmpdir))



import sys
from typing import NoReturn

import rich.console

ERROR_INVALID_PROJECT = 10


def exit_for_invalid_project(msg: str) -> NoReturn:
    # Be very helpful - this is likely the user's first interaction with chart-review.
    stderr = rich.console.Console(stderr=True)
    stderr.print(msg, style="bold red", highlight=False)
    stderr.print()
    stderr.print("This does not appear to be a chart-review project folder.")
    stderr.print("See https://docs.smarthealthit.org/cumulus/chart-review/ to set up your project.")
    stderr.print()
    stderr.print("Or pass --help for usage info.")
    sys.exit(ERROR_INVALID_PROJECT)



"""Tests for commands/mcnemar.py"""

from tests import base


class TestMcNemar(base.TestCase):
    """Test case for the McNemar code"""

    def test_default_output(self):
        stdout = self.run_cli(
            "mcnemar", "alice", "bob", "carla", path=f"{self.DATA_DIR}/many-notes"
        )

        # This tests the mid-P cutoff for b+c<25 (label B) as well as a low-significance and
        # a high-significance label.
        self.assertEqual(
            """Comparing 50 charts (1–50)
Truth: alice
Annotators: bob, carla

McNemar  P-value   BC  OL  OR  BW  Label
12.375   4.35e-04  20  61  27  42  *    
6.5      0.011     10  20  6   14  A    
N/A      0.035     1   16  6   27  B    
2.025    0.155     9   25  15  1   C    
""",
            stdout,
        )

    def test_csv(self):
        stdout = self.run_cli(
            "mcnemar", "--csv", "alice", "bob", "carla", path=f"{self.DATA_DIR}/many-notes"
        )

        self.assertEqual(
            [
                "mcnemar,p-value,bc,ol,or,bw,label",
                "12.375,4.35e-04,20,61,27,42,*",
                "6.5,0.011,10,20,6,14,A",
                ",0.035,1,16,6,27,B",
                "2.025,0.155,9,25,15,1,C",
            ],
            stdout.splitlines(),
        )

    def test_bad_annotator(self):
        """Verify that we do something suitable for bad arguments"""
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("mcnemar", "nope2", "john", "jill", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Unrecognized annotator 'nope2'\n", stderr.getvalue())

    def test_same_annotators(self):
        """Verify that we do something suitable for bad arguments"""
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("mcnemar", "john", "jill", "jill", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Can’t compare the same annotator with themselves.\n", stderr.getvalue())

    def test_sublabels(self):
        output = self.run_cli("mcnemar", "alice", "bob", "carla", path=f"{self.DATA_DIR}/sublabels")
        self.assertEqual(
            output,
            """Comparing 2 charts (1308–1309)
Truth: alice
Annotators: bob, carla

McNemar  P-value  BC  OL  OR  BW  Label                         
N/A      0.688    10  2   3   1   *                             
N/A      1.0      1   0   0   1   Deceased                      
N/A      1.0      2   0   0   0   Deceased → False              
N/A      0.5      1   0   1   0   Deceased → Datetime → 11/12/25
N/A      0.5      1   0   1   0   Deceased → Datetime → 11/13/25
N/A      1.0      2   0   0   0   Fungal → Confirmed            
N/A      0.5      1   0   1   0   Infection                     
N/A      0.5      1   1   0   0   Infection → Confirmed         
N/A      0.5      1   1   0   0   Infection → Suspected         
""",
        )



"""Tests for studio.py"""

import json
import tempfile

from chart_review import studio
from tests import base


class TestStudio(base.TestCase):
    """Test some edge cases with Label Studio export parsing"""

    def test_invalid_label_type(self):
        with tempfile.NamedTemporaryFile("wt") as tmpfile:
            json.dump(
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {
                                        "value": {"labels": ["A"]},
                                        "type": "bogus",
                                    }
                                ],
                            },
                        ],
                    },
                ],
                tmpfile,
            )
            tmpfile.flush()
            with self.assertRaisesRegex(ValueError, "Unrecognized Label Studio result type"):
                studio.ExportFile(tmpfile.name)

    def test_invalid_label_id(self):
        with tempfile.NamedTemporaryFile("wt") as tmpfile:
            json.dump(
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {
                                        "id": "id1",
                                        "value": {"labels": ["A"]},
                                        "type": "labels",
                                        "from_name": "label",
                                    },
                                    {
                                        "id": "id1",
                                        "value": {"choices": ["A1.1"]},
                                        "type": "choices",
                                        "from_name": "A",
                                    },
                                    {
                                        "id": "bogus-id",
                                        "value": {"choices": ["A1.2"]},
                                        "type": "choices",
                                        "from_name": "A",
                                    },
                                ],
                            },
                        ],
                        "data": {"text": "", "label": []},
                    },
                ],
                tmpfile,
            )
            tmpfile.flush()
            with self.assertRaisesRegex(ValueError, "Unrecognized sublabel ID 'bogus-id'."):
                studio.ExportFile(tmpfile.name)



"""Methods for showing config & calculated setup info."""

import argparse

import rich
import rich.box
import rich.table

import chart_review
from chart_review import cli_utils, console_utils


def print_info(args: argparse.Namespace) -> None:
    """Show project information on the console."""
    reader = cli_utils.get_cohort_reader(args)
    console = rich.get_console()

    # Charts
    chart_table = rich.table.Table(
        "Annotator",
        "Chart Count",
        "Chart IDs",
        box=rich.box.ROUNDED,
    )
    for annotator in sorted(reader.note_range):
        notes = reader.note_range[annotator]
        chart_table.add_row(
            annotator,
            f"{len(notes):,}",
            console_utils.pretty_note_range(notes),
        )

    console.print(chart_table)
    console_utils.print_ignored_charts(reader)

    console.print()
    console.print("Pass --help to see more options.")


def make_subparser(parser: argparse.ArgumentParser) -> None:
    cli_utils.add_project_args(parser, is_global=True)
    parser.add_argument(
        "--version",
        action="version",
        version=f"chart-review {chart_review.__version__}",
    )
    parser.set_defaults(func=print_info)



"""Tests for commands/mentions.py"""

import tempfile

from chart_review import common
from tests import base


class TestMentions(base.TestCase):
    """Test case for the top-level mentions code"""

    def test_basic_output(self):
        stdout = self.run_cli("mentions", path=f"{self.DATA_DIR}/cold")

        self.assertEqual(
            """╭───────────┬──────────┬─────────┬──────────╮
│ Annotator │ Chart ID │ Mention │ Label    │
├───────────┼──────────┼─────────┼──────────┤
│ jane      │ 1        │ achoo   │ Cough    │
│ jane      │ 1        │ sigh    │ Fatigue  │
│ jane      │ 1        │ sigh    │ Headache │
│ jane      │ 4        │ pain    │ Headache │
│ jane      │ 4        │ sigh    │ Fatigue  │
│ jane      │ 4        │ sleepy  │ Fatigue  │
├───────────┼──────────┼─────────┼──────────┤
│ jill      │ 1        │ achoo   │ Cough    │
│ jill      │ 1        │ sigh    │ Fatigue  │
│ jill      │ 2        │ ouch    │ Fatigue  │
│ jill      │ 4        │ pain    │ Cough    │
│ jill      │ 4        │ sleepy  │ Fatigue  │
├───────────┼──────────┼─────────┼──────────┤
│ john      │ 1        │ achoo   │ Cough    │
│ john      │ 1        │ sigh    │ Fatigue  │
│ john      │ 2        │ ouch    │ Headache │
│ john      │ 4        │ pain    │ Headache │
│ john      │ 4        │ sleepy  │ Fatigue  │
╰───────────┴──────────┴─────────┴──────────╯
""",
            stdout,
        )

    def test_ignored(self):
        """Verify that we show info on ignored notes"""
        stdout = self.run_cli("mentions", path=f"{self.DATA_DIR}/ignore")

        # Blank mentions are correct - the ignore project doesn't list the source text.
        # Good to confirm that we still do something reasonable in this edge case.
        self.assertEqual(
            """╭───────────┬──────────┬─────────┬───────╮
│ Annotator │ Chart ID │ Mention │ Label │
├───────────┼──────────┼─────────┼───────┤
│ adam      │ 1        │         │ A     │
│ adam      │ 2        │         │ B     │
├───────────┼──────────┼─────────┼───────┤
│ allison   │ 1        │         │ A     │
│ allison   │ 2        │         │ B     │
╰───────────┴──────────┴─────────┴───────╯
  Ignoring 3 charts (3–5)
""",
            stdout,
        )

    def test_external(self):
        """Verify that we don't show external annotators"""
        stdout = self.run_cli("mentions", path=f"{self.DATA_DIR}/external")

        self.assertEqual(
            """╭───────────┬──────────┬─────────┬───────╮
│ Annotator │ Chart ID │ Mention │ Label │
├───────────┼──────────┼─────────┼───────┤
│ human     │ 1        │ sigh    │ sad   │
│ human     │ 1        │ woo     │ happy │
╰───────────┴──────────┴─────────┴───────╯
""",
            stdout,
        )

    def test_odd_text(self):
        """Verify that unusual text like multi-word unicode doesn't trip us up"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(f"{tmpdir}/config.json", {"annotators": {"chris": 1}})
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {"value": {"text": "Cute Li🦁n", "labels": ["Cat"]}},
                                    {"value": {"text": "Multi\nLine-on", "labels": ["Cat"]}},
                                ],
                            },
                        ],
                    },
                ],
            )
            stdout = self.run_cli("mentions", path=tmpdir)

            self.assertEqual(
                """╭───────────┬──────────┬────────────┬───────╮
│ Annotator │ Chart ID │ Mention    │ Label │
├───────────┼──────────┼────────────┼───────┤
│ chris     │ 1        │ cute li🦁n │ Cat   │
│ chris     │ 1        │ multi      │ Cat   │
│           │          │ line-on    │       │
╰───────────┴──────────┴────────────┴───────╯
""",
                stdout,
            )

    def test_unused_labels(self):
        """Verify that we don't list mentions for labels that aren't in consideration"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "annotators": {"chris": 1},
                    "labels": ["Valid"],
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {"value": {"text": "good", "labels": ["Valid"]}},
                                    {"value": {"text": "bad", "labels": ["Invalid"]}},
                                ],
                            },
                        ],
                    },
                ],
            )
            stdout = self.run_cli("mentions", path=tmpdir)

            self.assertEqual(
                """╭───────────┬──────────┬─────────┬───────╮
│ Annotator │ Chart ID │ Mention │ Label │
├───────────┼──────────┼─────────┼───────┤
│ chris     │ 1        │ good    │ Valid │
╰───────────┴──────────┴─────────┴───────╯
""",
                stdout,
            )

    def test_duplicate_mention(self):
        """Verify that we don't show two copies of the same information"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "annotators": {"chris": 1},
                    "labels": ["LabelA", "LabelB"],
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {"value": {"text": "dup", "labels": ["LabelA"]}},
                                    {"value": {"text": "dup", "labels": ["LabelA"]}},
                                    {"value": {"text": "new", "labels": ["LabelA"]}},
                                    {"value": {"text": "new", "labels": ["LabelB"]}},
                                ],
                            },
                        ],
                    },
                ],
            )
            stdout = self.run_cli("mentions", path=tmpdir)

            self.assertEqual(
                """╭───────────┬──────────┬─────────┬────────╮
│ Annotator │ Chart ID │ Mention │ Label  │
├───────────┼──────────┼─────────┼────────┤
│ chris     │ 1        │ dup     │ LabelA │
│ chris     │ 1        │ new     │ LabelA │
│ chris     │ 1        │ new     │ LabelB │
╰───────────┴──────────┴─────────┴────────╯
""",
                stdout,
            )

    def test_csv(self):
        """Verify that can print in CSV format"""
        stdout = self.run_cli("mentions", "--csv", path=f"{self.DATA_DIR}/external")

        self.assertEqual(
            [
                "annotator,chart_id,mention,label",
                "human,1,sigh,sad",
                "human,1,woo,happy",
            ],
            stdout.splitlines(),
        )

    def test_sublabels(self):
        output = self.run_cli("mentions", path=f"{self.DATA_DIR}/sublabels")
        self.assertEqual(
            output,
            """╭───────────┬──────────┬───────────────────────┬───────────────────────────────╮
│ Annotator │ Chart ID │ Mention               │ Label                         │
├───────────┼──────────┼───────────────────────┼───────────────────────────────┤
│ alice     │ 1308     │ alive                 │ Deceased → False              │
│ alice     │ 1308     │ alive                 │ Deceased → Datetime →         │
│           │          │                       │ 11/12/25                      │
│ alice     │ 1308     │ death                 │ Deceased                      │
│ alice     │ 1308     │ fungus found in lungs │ Fungal → Confirmed            │
│ alice     │ 1308     │ maybe got infected    │ Infection → Suspected         │
├───────────┼──────────┼───────────────────────┼───────────────────────────────┤
│ bob       │ 1308     │ alive                 │ Deceased → False              │
│ bob       │ 1308     │ alive                 │ Deceased → Datetime →         │
│           │          │                       │ 11/13/25                      │
│ bob       │ 1308     │ death                 │ Infection                     │
│ bob       │ 1308     │ fungus found in lungs │ Fungal → Confirmed            │
│ bob       │ 1308     │ maybe got infected    │ Infection → Suspected         │
├───────────┼──────────┼───────────────────────┼───────────────────────────────┤
│ carla     │ 1308     │ alive                 │ Deceased → False              │
│ carla     │ 1308     │ alive                 │ Deceased → Datetime →         │
│           │          │                       │ 11/12/25                      │
│ carla     │ 1308     │ fungus found in lungs │ Fungal → Confirmed            │
│ carla     │ 1308     │ maybe got infected    │ Infection → Confirmed         │
╰───────────┴──────────┴───────────────────────┴───────────────────────────────╯
""",
        )



"""Match external document references & labels to Label Studio data"""

import csv
import os
import sys

from chart_review import defines, studio


class ExternalCsvParser:
    def __init__(self, filename: str):
        self.filename = filename
        self.file = None

    def __enter__(self):
        self.file = open(self.filename, newline="", encoding="utf8")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()

    def __iter__(self):
        self._read_headers()
        return self

    def __next__(self) -> tuple[str, defines.Label | None]:
        row = next(self.reader)
        return self._row_to_id(row), self._row_to_label(row)

    def _read_headers(self):
        self.reader = csv.reader(self.file)

        # Remove case concerns
        header = next(self.reader, None)
        header = [x.casefold() for x in header]

        self.sublabel_name_col = None
        self.sublabel_value_col = None

        # There are two ways headers could be layed out:
        # - bare bones two column layout of id/label
        # - an ID column (with lots of possible names) and the standard label column names

        if len(header) == 2:
            # Implicit header order of [id, label]
            self.id_col = 0
            self.label_col = 1
            # The ID could be any of DocumentReference, DiagnosticReport, or Encounter.
            self.default_resource = self._check_col_name_for_res(header[0])
            if not self.default_resource:
                print(
                    f"Unrecognized ID column '{header[0]}'. Will assume Encounter ID.",
                    file=sys.stderr,
                )
                self.default_resource = "Encounter"
        else:
            # Search for ID column, assume label cols are using standard names
            for index, col_name in enumerate(header):
                self.default_resource = self._check_col_name_for_res(col_name)
                if self.default_resource:
                    self.id_col = index
                    break
            else:
                self._error("no resource ID column found")
            try:
                self.label_col = header.index("label")
            except ValueError:
                self._error("no 'label' column found")
            try:
                self.sublabel_name_col = header.index("sublabel_name")
            except ValueError:
                pass  # this is allowed
            if self.sublabel_name_col is not None:
                try:
                    self.sublabel_value_col = header.index("sublabel_value")
                except ValueError:
                    self._error("no 'sublabel_value' column found")

    def _row_to_id(self, row: list[str]) -> str:
        row_id = row[self.id_col]
        if "/" not in row_id:
            row_id = f"{self.default_resource}/{row_id}"
        return row_id

    def _row_to_label(self, row: list[str]) -> defines.Label | None:
        if not row[self.label_col]:  # allow for no labels for a row (no positive labels found)
            return None

        args = [row[self.label_col]]
        if self.sublabel_name_col is not None and self.sublabel_value_col is not None:
            args.append(row[self.sublabel_name_col])
            args.append(row[self.sublabel_value_col])

        return defines.Label(*args)

    def _check_col_name_for_res(self, col_name: str) -> str | None:
        if "doc" in col_name or col_name == "note_ref":
            return "DocumentReference"
        elif col_name in {"diagnosticreport_id", "diagnosticreport_ref"}:
            return "DiagnosticReport"
        elif "enc" in col_name:
            return "Encounter"
        else:
            return None

    def _error(self, msg: str):
        raise ValueError(f"Could not parse external file '{self.filename}': {msg}.")


def _load_csv_labels(filename: str) -> dict[str, defines.LabelSet]:
    """
    Loads a csv and returns a list of labels per row.

    CSV format is two columns, where the first is note/encounter id and the second is a single
    label.

    Returns {row_id -> set of labels for that ID}
    """
    id_to_labels = {}

    with ExternalCsvParser(filename) as parser:
        for row_id, label in parser:
            label_set = id_to_labels.setdefault(row_id, defines.LabelSet())
            if label:  # can be None if no labels given for a row
                label_set.add(label)

    return id_to_labels


def _note_ref_to_label_studio_id(export: studio.ExportFile, note_ref: str) -> int | None:
    """Looks at the metadata in LS and grabs the note ID that holds the provided note ref"""
    for note in export.notes:
        for key, value in note.docref_mappings.items():
            # Support older exports that didn't specify DocRef vs DxReport
            key = key if "/" in key else f"DocumentReference/{key}"
            value = value if "/" in value else f"DocumentReference/{value}"
            # Allow either an anonymous ID or the real ID -- collisions seem very unlikely
            # (i.e. real IDs aren't going to be formatted like our long anonymous ID hash)
            if key == note_ref or value == note_ref:
                return note.note_id
    return None


def _encounter_id_to_label_studio_id(export: studio.ExportFile, enc_id: str) -> int | None:
    """Looks at the metadata in LS and grabs the note ID that holds the provided encounter"""
    for note in export.notes:
        # Allow either an anonymous ID or the real ID -- collisions seem very unlikely
        # (i.e. real IDs aren't going to be formatted like our long anonymous ID hash)
        if note.encounter_id == enc_id or note.anon_encounter_id == enc_id:
            return note.note_id
    return None


def external_id_to_label_studio_id(
    export: studio.ExportFile,
    row_id: str,
) -> int | None:
    """Looks at the metadata in LS and grabs the note ID that holds the provided ID"""
    # First, check if there is a resource prefix, which will tell us which kind of ID this is
    parts = row_id.split("/", 1)
    if parts[0] == "Encounter" or len(parts) == 1:
        return _encounter_id_to_label_studio_id(export, parts[-1])
    elif parts[0] in {"DiagnosticReport", "DocumentReference"}:
        return _note_ref_to_label_studio_id(export, row_id)
    else:
        raise ValueError(f"Unrecognized resource type: {parts[0]}")  # pragma: no cover


def merge_external(
    annotations: defines.ProjectAnnotations,
    export: studio.ExportFile,
    project_dir: str,
    name: str,
    config: dict,
) -> None:
    """Loads an external csv file annotator and merges them into an existing simple dict"""
    if isinstance(config, dict) and (filename := config.get("filename")):
        full_filename = os.path.join(project_dir, filename)
        label_map = _load_csv_labels(full_filename)
    else:
        raise ValueError(f"Did not understand config for external annotator '{name}'")

    # Inspect exported json to see if it has the metadata we'll need.
    for note in export.notes:
        if not note.docref_mappings:
            raise ValueError(
                "Your Label Studio export does not include note/encounter ID mapping metadata!\n"
                "Consider re-uploading your notes using Cumulus ETL's upload-notes command."
            )
        break  # just inspect one

    # Convert each row id into an LS id:
    external_mentions = annotations.mentions.setdefault(name, defines.Mentions())
    for row_id, label_set in label_map.items():
        ls_id = external_id_to_label_studio_id(export, row_id)
        if ls_id is not None:
            all_labels = external_mentions.setdefault(ls_id, set())
            all_labels |= label_set



"""Various type declarations for better type hinting."""

import dataclasses
import functools


def _split_label(label_str: str) -> tuple[str, str, str]:
    pieces = label_str.split("|", 2)
    while len(pieces) < 3:
        pieces.append("")
    return tuple(piece.strip() for piece in pieces)


@dataclasses.dataclass(frozen=True)
@functools.total_ordering
class Label:
    label: str
    sublabel_name: str = ""  # empty string means "not provided" / "not used"
    sublabel_value: str = ""

    @classmethod
    def parse(cls, label_str: str) -> "Label":
        """
        This parses a label string into a full Label class.

        It interprets '|' characters as delimiters for the label and sublabel name.
        Surrounding whitespace of any part of the label string is ignored.
        """
        return Label(*_split_label(label_str))

    def __post_init__(self):
        if "|" in self.label or "|" in self.sublabel_name:
            raise ValueError("Invalid character found in label name: '|'.")
        if self.sublabel_name and not self.sublabel_value:
            raise ValueError(
                f"Sublabel name but no sublabel value provided: '{self.sublabel_name}'."
            )

    def __str__(self) -> str:
        """Suitable for presenting to user, though it may be long"""
        if self.sublabel_name:
            short_name = self.sublabel_name.removeprefix(f"{self.label} ")
            if not short_name or self.label == self.sublabel_name:
                return f"{self.label} → {self.sublabel_value}"
            else:
                return f"{self.label} → {short_name} → {self.sublabel_value}"
        else:
            return self.label

    def __lt__(self, other) -> bool:
        """Case-insensitive ordering (useful for presentation to user)"""
        left = tuple(x.casefold() for x in dataclasses.astuple(self))
        right = tuple(x.casefold() for x in dataclasses.astuple(other))
        return left < right

    def namespace(self) -> tuple[str, str]:
        """
        Returns an object combining the "namespace" (i.e. "non-value") parts of the label.

        When sublabels are present, it's the label and sublabel, but not the sublabel value.
        When not using a sublabel, it's just empty strings because the label itself is a value in
        the same "namespace" as all other bare labels.

        This is mostly used for detecting when two labels "conflict", meaning they have the same
        namespace but different values. (like when `frequency` detects conflicting labels for the
        same text)
        """
        if self.sublabel_name:
            return self.label, self.sublabel_name
        else:
            return "", ""


# Map of label_studio_user_id: human name
AnnotatorMap = dict[int, str]

LabelSet = set[Label]
NoteSet = set[int]

# Map of label_studio_note_id: {all labels for that note}
# Usually used in the context of a specific annotator's label mentions.
Mentions = dict[int, LabelSet]


class LabelMatcher:
    """
    Class to match a specific label or a class of labels.

    Examples:
        "A|B|C" will match "A|B|C" but not "A|B|D"
        "A|B" will match both "A|B|C" and "A|B|D" but not "A|E"
        "A" will match "A|B|C" and "A|E|F" and "A" but not "X"
    """

    def __init__(self, *expressions: str):
        self._labels = frozenset(_split_label(x) for x in expressions)

    def __eq__(self, other):
        return self._labels == other._labels

    def __hash__(self):
        return hash(self._labels)

    def is_match(self, other: Label) -> bool:
        for label in self._labels:
            if (
                label[0] == other.label
                and (not label[1] or label[1] == other.sublabel_name)
                and (not label[2] or label[2] == other.sublabel_value)
            ):
                return True
        return False

    def matches_in_set(self, other: LabelSet) -> LabelSet:
        return {label for label in other if self.is_match(label)}


# Map of label: {all implied labels}
ImpliedLabels = dict[LabelMatcher, LabelSet]

# Map of group: {all member labels}
GroupedLabels = dict[Label, LabelMatcher]


@dataclasses.dataclass
class LabeledText:
    text: str | None
    labels: LabelSet


@dataclasses.dataclass
class ProjectAnnotations:
    labels: LabelSet = dataclasses.field(default_factory=set)

    # annotator_name -> Mentions
    mentions: dict[str, Mentions] = dataclasses.field(default_factory=dict)

    # We usually deal with simplified note-wide labels, but sometimes it's helpful to have
    # the original text-to-label associations available, for term frequency analysis.
    # annotators -> note_id -> list of text/label combos
    original_text_mentions: dict[str, dict[int, list[LabeledText]]] = dataclasses.field(
        default_factory=dict
    )

    def remove(self, chart_id: int):
        # Remove any instance of this chart ID
        for mentions in self.mentions.values():
            if chart_id in mentions:
                del mentions[chart_id]
        for mentions in self.original_text_mentions.values():
            if chart_id in mentions:
                del mentions[chart_id]



"""Tests for cohort.py"""

import tempfile

from chart_review import cohort, common, config
from tests import base


class TestCohort(base.TestCase):
    """Test case for basic cohort management"""

    def test_no_specified_label(self):
        """Verify that no label setup grabs all found labels from the export."""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "annotators": {"bob": 1, "alice": 2},
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {
                                        "value": {"labels": ["Label A", "Label B"]},
                                    }
                                ],
                            },
                        ],
                    },
                ],
            )
            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))

        self.assertEqual(base.labels({"Label A", "Label B"}), reader.class_labels)

    def test_ignored_ids(self):
        reader = cohort.CohortReader(config.ProjectConfig(f"{self.DATA_DIR}/ignore"))

        # Confirm 3, 4, and 5 got ignored
        self.assertEqual(
            {
                "adam": {1, 2},
                "allison": {1, 2},
            },
            reader.note_range,
        )

    def test_non_existent_ids(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json", {"annotators": {"bob": 1}, "ranges": {"bob": ["1-5"]}}
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {"id": 1, "annotations": [{"completed_by": 1}]},  # done by bob
                    {"id": 3},  # not done by bob, but we are explicitly told it was
                ],
            )
            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))

            self.assertEqual({"bob": {1, 3}}, reader.note_range)

    def test_unknown_annotators_are_ignored(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(f"{tmpdir}/config.json", {"annotators": {"bob": 1}})
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {"id": 1, "annotations": [{"completed_by": 1}, {"completed_by": 2}]},
                ],
            )
            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))
            self.assertEqual({"bob": {1}}, reader.note_range)

    def test_uncredited_annotations_are_ignored(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [{"value": {"labels": ["cat"]}}],
                            },
                            {
                                # No completed_by
                                "result": [{"value": {"labels": ["bear"]}}],
                            },
                        ],
                    },
                ],
            )
            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))
            self.assertEqual({"1": {1: base.labels({"cat"})}}, reader.annotations.mentions)

    def test_default_annotator_config(self):
        """Should use a string version of the completed_by ID for the names"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {"id": 1, "annotations": [{"completed_by": 1}, {"completed_by": 2}]},
                ],
            )
            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))
            self.assertEqual({"1": {1}, "2": {1}}, reader.note_range)

    def test_implied_labels_get_expanded(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "annotators": {"bob": 1},
                    "implied-labels": {
                        "cat": ["animal", "cat"],
                        "animal": "alive",
                    },
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {"completed_by": 1, "result": [{"value": {"labels": ["cat"]}}]},
                        ],
                    }
                ],
            )
            reader = cohort.CohortReader(config.ProjectConfig(tmpdir))
            self.assertEqual(
                {"bob": {1: base.labels({"cat", "animal", "alive"})}}, reader.annotations.mentions
            )



import argparse

import rich.table

from chart_review import cli_utils


def make_subparser(parser: argparse.ArgumentParser) -> None:
    cli_utils.add_project_args(parser)
    cli_utils.add_output_args(parser)
    parser.set_defaults(func=print_ids)


def print_ids(args: argparse.Namespace) -> None:
    """
    Prints a mapping of all project IDs.

    Currently, this writes a CSV file to stdout. In the future, this could get fancier.
    At the time of writing, it wasn't clear how to present the information in a way that
    sensible to a casual console user - so I went with the more technical-oriented CSV file.
    """
    reader = cli_utils.get_cohort_reader(args)

    table = cli_utils.create_table("Chart ID", "Original FHIR ID", "Anonymized FHIR ID")

    for note in reader.ls_export.notes:
        chart_id = str(note.note_id)
        printed = False

        # Grab encounters first
        orig_id = note.encounter_id and f"Encounter/{note.encounter_id}"
        anon_id = note.anon_encounter_id and f"Encounter/{note.anon_encounter_id}"
        if orig_id or anon_id:
            table.add_row(chart_id, orig_id, anon_id)
            printed = True

        # Now each DocRef ID
        for orig_id, anon_id in note.docref_mappings.items():
            table.add_row(chart_id, f"DocumentReference/{orig_id}", f"DocumentReference/{anon_id}")
            printed = True

        if not printed:
            # Guarantee that every Chart ID shows up at least once - so it's clearer that the
            # chart ID is included in the Label Studio export but that it does not have any
            # IDs mapped to it.
            table.add_row(chart_id, None, None)

    if args.csv:
        cli_utils.print_table_as_csv(table)
    else:
        rich.get_console().print(table)



from collections.abc import Collection, Iterable

from chart_review import defines


def confusion_matrix(
    annotations: defines.ProjectAnnotations,
    truth: str,
    annotator: str,
    note_range: Collection[int],
    labels: Iterable[str] | None = None,
) -> dict[str, list]:
    """
    Confusion Matrix (TP, FP, TN, FN)
    https://www.researchgate.net/figure/Calculation-of-sensitivity-specificity-and-positive-and-negative-predictive_fig1_49650721

    This is the rollup of counting each symptom only once, not multiple times for a single patient.
    :param annotations: prepared map of annotators & mentions
    :param truth: annotator to use as the ground truth
    :param annotator: another annotator to compare with truth
    :param note_range: collection of LabelStudio document ID
    :param labels: (optional) collection of labels to consider examining
    :return: Dict
        "TP": True Positives (agree on positive+ symptom)
        "FP": False Positives (annotator said positive+, truth said No)
        "FN": False Negative (truth said positive+, annotator said No)
        "TN": True Negative (truth and annotator both said No)
    """
    truth_mentions = annotations.mentions.get(truth, defines.Mentions())
    annotator_mentions = annotations.mentions.get(annotator, defines.Mentions())

    # Only examine labels that were used by any compared annotators at least once
    label_set = set()
    for _v in truth_mentions.values():
        label_set |= set(_v)
    for _v in annotator_mentions.values():
        label_set |= set(_v)
    if labels:
        label_set &= set(labels)

    TP = list()  # True Positive
    FP = list()  # False Positive
    FN = list()  # False Negative
    TN = list()  # True Negative

    for note_id in note_range:
        truth_note_mentions = truth_mentions.get(note_id, set())
        annotator_note_mentions = annotator_mentions.get(note_id, set())

        for label in sorted(label_set):
            key = {note_id: label}
            truth_positive = label in truth_note_mentions
            annotator_positive = label in annotator_note_mentions

            if truth_positive and annotator_positive:
                TP.append(key)
            elif truth_positive and not annotator_positive:
                FN.append(key)
            elif not truth_positive and annotator_positive:
                FP.append(key)
            elif not truth_positive and not annotator_positive:
                TN.append(key)
            else:
                raise Exception("Guard: Impossible comparison of reviewers")  # pragma: no cover

    return {"TP": TP, "FN": FN, "FP": FP, "TN": TN}


def score_kappa(matrix: dict) -> float:
    """
    Computes Cohen kappa for pair-wise annotators.
    https://en.wikipedia.org/wiki/Cohen%27s_kappa

    :param matrix: confusion matrix with TN/TP/FN/FP values
    :return: Cohen kappa statistic
    """
    tp = len(matrix["TP"])  # true positive
    tn = len(matrix["TN"])  # true negative
    fp = len(matrix["FP"])  # false positive
    fn = len(matrix["FN"])  # false negative
    total = tp + tn + fp + fn

    # observed agreement A (Po)
    observed = (tp + tn) / total

    # expected agreement E (Pe)
    expected_pos = ((tp + fp) / total) * ((tp + fn) / total)
    expected_neg = ((tn + fp) / total) * ((tn + fn) / total)
    expected = expected_pos + expected_neg

    return (observed - expected) / (1 - expected)


def score_matrix(matrix: dict) -> dict:
    """
    Score F1 and Kappa measures with precision (PPV) and recall (sensitivity).
    F1 deliberately ignores "True Negatives" because TN inflates scoring (AUROC)
    @return: dict with keys {'f1', 'precision', 'recall'} vals are %score
    """
    true_pos = len(matrix["TP"])
    true_neg = len(matrix["TN"])
    false_pos = len(matrix["FP"])
    false_neg = len(matrix["FN"])

    if 0 == true_pos or 0 == true_neg:
        sens = 0
        spec = 0
        ppv = 0
        npv = 0
        f1 = 0
        kappa = 0
    else:
        sens = true_pos / (true_pos + false_neg)
        spec = true_neg / (true_neg + false_pos)
        ppv = true_pos / (true_pos + false_pos)
        npv = true_neg / (true_neg + false_neg)
        f1 = (2 * ppv * sens) / (ppv + sens)
        kappa = score_kappa(matrix)

    return {
        "F1": f1,
        "Sens": sens,
        "Spec": spec,
        "PPV": ppv,
        "NPV": npv,
        "Kappa": kappa,
        "TP": true_pos,
        "FP": false_pos,
        "FN": false_neg,
        "TN": true_neg,
    }


def csv_table(score: dict, class_labels: defines.LabelSet):
    table = list()
    table.append(csv_header(False, True))
    table.append(csv_row_score(score, as_string=True))

    for label in sorted(class_labels):
        table.append(csv_row_score(score[label], str(label), as_string=True))
    return "\n".join(table) + "\n"


def csv_header(pick_label=False, as_string=False):
    """
    Table Header
    F1, PPV (precision), Recall (sensitivity), True Pos, False Pos, False Neg
    :param pick_label: default= None
    :return: header
    """
    as_list = ["F1", "Sens", "Spec", "PPV", "NPV", "Kappa", "TP", "FN", "TN", "FP"]

    if not as_string:
        return as_list

    header = as_list
    header.append(pick_label if pick_label else "Label")
    return "\t".join(header)


def csv_row_score(
    score: dict, pick_label: str | None = None, as_string: bool = False
) -> str | list[str]:
    """
    Table Row entry
    F1, PPV (precision), Recall (sensitivity), True Pos, False Pos, False Neg
    :param score: dict result from F1 scoring
    :param pick_label: default= None means '*' all classes
    :param as_string: whether to return a list of string scores or one single string
    :return: str representation of the score
    """
    row = [score[header] for header in csv_header()]
    row = [str(round(value, 3)) for value in row]

    if not as_string:
        return row

    row.append(pick_label if pick_label else "*")
    return "\t".join(row)


def contingency_table(
    annotations: defines.ProjectAnnotations,
    truth: str,
    annotator1: str,
    annotator2: str,
    note_range: Collection[int],
    labels: Iterable[str] | None = None,
) -> dict[str, list]:
    # Grab all mentions
    mentions = {
        annotator: annotations.mentions.get(annotator, defines.Mentions())
        for annotator in {truth, annotator1, annotator2}
    }

    # Only examine labels that were used by any compared annotators at least once
    label_set = set()
    for mentions_set in mentions.values():
        for mention_labels in mentions_set.values():
            label_set |= set(mention_labels)
    if labels:
        label_set &= set(labels)

    BC = []  # both correct
    OL = []  # only left
    OR = []  # only right
    BW = []  # both wrong

    for note_id in note_range:
        truth_note_mentions = mentions[truth].get(note_id, set())

        for label in sorted(label_set):
            key = {note_id: label}

            correctness = []
            for annotator in [annotator1, annotator2]:
                annotator_note_mentions = mentions[annotator].get(note_id, set())

                truth_positive = label in truth_note_mentions
                annotator_positive = label in annotator_note_mentions

                correctness.append(truth_positive == annotator_positive)

            if correctness == [True, True]:
                BC.append(key)
            elif correctness == [True, False]:
                OL.append(key)
            elif correctness == [False, True]:
                OR.append(key)
            elif correctness == [False, False]:
                BW.append(key)
            else:
                raise Exception("Guard: Impossible comparison of reviewers")  # pragma: no cover

    return {"BC": BC, "OL": OL, "OR": OR, "BW": BW}



"""Base class for tests"""

import contextlib
import io
import os
import unittest
from collections.abc import Iterable

from chart_review import cli, defines

Label = defines.Label  # syntactic sugar
LabelMatcher = defines.LabelMatcher  # syntactic sugar


def labels(strs: Iterable[str]) -> defines.LabelSet:
    return {defines.Label.parse(x) for x in strs}


class TestCase(unittest.TestCase):
    """Test case parent class"""

    DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

    def setUp(self):
        super().setUp()
        self.maxDiff = None

    @staticmethod
    def run_cli(*args, path: str) -> str:
        args = ["--project-dir", path, *args]
        with TestCase.capture_stdout() as stdout:
            cli.main_cli(args)
        return stdout.getvalue()

    @staticmethod
    @contextlib.contextmanager
    def capture_stdout() -> io.StringIO:
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            yield stdout

    @staticmethod
    @contextlib.contextmanager
    def capture_stderr() -> io.StringIO:
        stderr = io.StringIO()
        with contextlib.redirect_stderr(stderr):
            yield stderr



from chart_review import config, defines, studio


def simplify_export(
    export: studio.ExportFile, proj_config: config.ProjectConfig
) -> defines.ProjectAnnotations:
    """
    Label Studio outputs contain more info than needed for IAA and term_freq.

    * PHI raw physician note text is removed *

    :param export: exported json from Label Studio
    :param proj_config: project configuration
    :return: all project mentions parsed from the Label Studio export
    """
    annotations = defines.ProjectAnnotations()
    annotations.labels = proj_config.class_labels
    grab_all_labels = not annotations.labels

    for note in export.notes:
        for annot in note.annotations:
            if proj_config.annotators and annot.author not in proj_config.annotators:
                continue  # user specified an annotators config, and this one doesn't fit
            annotator = proj_config.annotators.get(annot.author, str(annot.author))

            # Grab all valid mentions for this annotator & note
            labels = defines.LabelSet()
            text_tags = []
            for mention in annot.mentions:
                labels |= mention.labels
                text_tags.append(defines.LabeledText(mention.text, mention.labels))

            if grab_all_labels:
                annotations.labels |= labels

            # Store these mentions in the main annotations list, by author & note
            annotator_mentions = annotations.mentions.setdefault(annotator, defines.Mentions())
            annotator_mentions[note.note_id] = labels
            annot_orig_text_tags = annotations.original_text_mentions.setdefault(annotator, {})
            annot_orig_text_tags[note.note_id] = text_tags

    return annotations


def _find_implied_labels(
    source_label: defines.Label,
    implied_label_mappings: defines.ImpliedLabels,
    found_labels: defines.LabelSet | None = None,
) -> defines.LabelSet:
    """
    Expands the source label into the set of all implied labels.

    Don't bother passing found_labels in, that's just a helper arg for recursion.
    """
    if found_labels is None:
        found_labels = set()

    if source_label in found_labels:
        return found_labels

    found_labels.add(source_label)
    for matcher, implied_labels in implied_label_mappings.items():
        if matcher.is_match(source_label):
            for implied_label in implied_labels:
                _find_implied_labels(
                    implied_label, implied_label_mappings, found_labels=found_labels
                )

    return found_labels


def _find_implied_mentions(
    mentions: defines.Mentions, implied_label_mappings: defines.ImpliedLabels
) -> defines.Mentions:
    """
    For every note, expands its labels into the set of all implied labels for that note.
    """
    found_mentions = defines.Mentions()

    for note_id, labels in mentions.items():
        implied_mentions = found_mentions.setdefault(note_id, set())
        for label in labels:
            implied_mentions |= _find_implied_labels(label, implied_label_mappings)

    return found_mentions


def _convert_grouped_mentions(
    mentions: defines.Mentions, grouped_label_mappings: defines.GroupedLabels
) -> defines.Mentions:
    """
    For every note, converts all labels in a group into one label for the group name.

    This is not recursive. (i.e. you can't have complicated grouping configs that combine)
    """
    final_mentions = defines.Mentions()

    for note_id, labels in mentions.items():
        for group_label, group_matcher in grouped_label_mappings.items():
            if matched := group_matcher.matches_in_set(labels):
                labels -= matched
                labels.add(group_label)
        final_mentions[note_id] = labels

    return final_mentions


def simplify_mentions(
    annotations: defines.ProjectAnnotations,
    *,
    implied_labels: defines.ImpliedLabels,
    grouped_labels: defines.GroupedLabels,
) -> None:
    # ** Expand all implied labels.
    annotations.mentions = {
        annotator: _find_implied_mentions(mentions, implied_labels)
        for annotator, mentions in annotations.mentions.items()
    }

    # ** Convert all grouped labels.
    # First, calculate the new set of valid labels, adding the group and removing the groupees
    annotations.labels |= set(grouped_labels.keys())
    annotations.labels.difference_update(
        *[m.matches_in_set(annotations.labels) for m in grouped_labels.values()]
    )
    # Next, convert old labels to the new group labels
    annotations.mentions = {
        annotator: _convert_grouped_mentions(mentions, grouped_labels)
        for annotator, mentions in annotations.mentions.items()
    }



"""Tests for cli.py"""

import shutil
import tempfile

import chart_review
from chart_review import cli, common, errors
from tests import base


class TestCommandLine(base.TestCase):
    """Test case for the CLI entry point"""

    def assert_cold_output(self, output):
        self.assertEqual(
            """╭───────────┬─────────────┬───────────╮
│ Annotator │ Chart Count │ Chart IDs │
├───────────┼─────────────┼───────────┤
│ jane      │ 3           │ 1, 3–4    │
│ jill      │ 4           │ 1–4       │
│ john      │ 3           │ 1–2, 4    │
╰───────────┴─────────────┴───────────╯

Pass --help to see more options.
""",
            output,
        )

    def test_version(self):
        # Manually capture stdout (rather than helper self.run_cli) because --version actually
        # exits the program, and we have to handle the exception rather than just grabbing the
        # return value which would be more convenient.
        with self.capture_stdout() as stdout:
            with self.assertRaises(SystemExit) as cm:
                cli.main_cli(["--version"])

        self.assertEqual(0, cm.exception.code)

        version = chart_review.__version__
        self.assertEqual(f"chart-review {version}\n", stdout.getvalue())

    def test_default_info(self):
        stdout = self.run_cli(path=f"{self.DATA_DIR}/cold")
        self.assert_cold_output(stdout)

    def test_default_info_ignored(self):
        stdout = self.run_cli(path=f"{self.DATA_DIR}/ignore")

        self.assertEqual(
            """╭───────────┬─────────────┬───────────╮
│ Annotator │ Chart Count │ Chart IDs │
├───────────┼─────────────┼───────────┤
│ adam      │ 2           │ 1–2       │
│ allison   │ 2           │ 1–2       │
╰───────────┴─────────────┴───────────╯
  Ignoring 3 charts (3–5)

Pass --help to see more options.
""",
            stdout,
        )

    def test_custom_config(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.copy(f"{self.DATA_DIR}/cold/labelstudio-export.json", tmpdir)
            # mostly confirm it doesn't just error out
            stdout = self.run_cli(f"--config={self.DATA_DIR}/cold/config.yaml", path=tmpdir)
            self.assert_cold_output(stdout)

    def test_missing_config(self):
        """Still works, just with rough defaults"""
        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.copy(f"{self.DATA_DIR}/cold/labelstudio-export.json", tmpdir)
            stdout = self.run_cli(path=tmpdir)

            self.assertEqual(
                """╭───────────┬─────────────┬───────────╮
│ Annotator │ Chart Count │ Chart IDs │
├───────────┼─────────────┼───────────┤
│ 3         │ 4           │ 1–4       │
│ 5         │ 3           │ 1–2, 4    │
│ 6         │ 4           │ 1–4       │
╰───────────┴─────────────┴───────────╯

Pass --help to see more options.
""",
                stdout,
            )

    def test_bad_config(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.copy(f"{self.DATA_DIR}/cold/labelstudio-export.json", tmpdir)
            common.write_text(f"{tmpdir}/config.json", "[1, 2]")
            with self.assertRaises(SystemExit) as cm:
                self.run_cli(path=tmpdir)
        self.assertEqual(cm.exception.code, errors.ERROR_INVALID_PROJECT)

    def test_missing_export(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(f"{tmpdir}/config.yaml", {"annotators": {"me": 1}})
            with self.assertRaises(SystemExit) as cm:
                self.run_cli(path=tmpdir)
        self.assertEqual(cm.exception.code, errors.ERROR_INVALID_PROJECT)



"""Helper methods for printing to the console."""

import rich

from chart_review import cohort, defines


def pretty_note_range(notes: defines.NoteSet) -> str:
    """
    Returns a pretty, human-readable string for a set of notes.

    If no notes, this returns an empty string.

    Example:
        pretty_note_range({1, 2, 3, 7, 9, 10})
        -> "1–3, 7, 9–10"
    """
    ranges = []
    range_start = None
    prev_note = None

    def end_range() -> None:
        if prev_note is None:
            return
        if range_start == prev_note:
            ranges.append(str(prev_note))
        else:
            ranges.append(f"{range_start}–{prev_note}")  # en dash

    for note in sorted(notes):
        if prev_note is None or prev_note + 1 != note:
            end_range()
            range_start = note
        prev_note = note

    end_range()

    return ", ".join(ranges)


def print_ignored_charts(reader: cohort.CohortReader):
    """
    Prints a line about ignored charts, suitable for underlying a table.

    It's recommended that any CLI command that shows individual chart IDs
    call this for their normal output view (i.e. not a formatted view like --csv).

    For commands that just show aggregate chart numbers,
    use your judgement if it helps or is just confusing extra info.
    """
    if not reader.ignored_notes:
        return

    ignored_count = len(reader.ignored_notes)
    chart_word = "chart" if ignored_count == 1 else "charts"
    pretty_ranges = pretty_note_range(reader.ignored_notes)
    rich.get_console().print(
        f"  Ignoring {ignored_count} {chart_word} ({pretty_ranges})",
        highlight=False,
        style="italic",
    )



"""Tests for commands/frequency.py"""

import tempfile

from chart_review import common
from tests import base


class TestFrequency(base.TestCase):
    """Test case for the top-level frequency code"""

    def test_basic_output(self):
        stdout = self.run_cli("frequency", path=f"{self.DATA_DIR}/cold")

        self.assertEqual(
            """╭───────────┬──────────┬─────────┬───────╮
│ Annotator │ Label    │ Mention │ Count │
├───────────┼──────────┼─────────┼───────┤
│ All       │ Cough    │ achoo   │ 3     │
│ All       │ Cough    │ pain*   │ 1     │
│ All       │ Fatigue  │ sigh*   │ 5     │
│ All       │ Fatigue  │ sleepy  │ 3     │
│ All       │ Fatigue  │ ouch*   │ 2     │
│ All       │ Headache │ pain*   │ 2     │
│ All       │ Headache │ sigh*   │ 1     │
│ All       │ Headache │ ouch*   │ 1     │
├───────────┼──────────┼─────────┼───────┤
│ jane      │ Cough    │ achoo   │ 1     │
│ jane      │ Fatigue  │ sigh*   │ 2     │
│ jane      │ Fatigue  │ sleepy  │ 1     │
│ jane      │ Headache │ sigh*   │ 1     │
│ jane      │ Headache │ pain*   │ 1     │
├───────────┼──────────┼─────────┼───────┤
│ jill      │ Cough    │ pain*   │ 1     │
│ jill      │ Cough    │ achoo   │ 1     │
│ jill      │ Fatigue  │ ouch*   │ 2     │
│ jill      │ Fatigue  │ sleepy  │ 1     │
│ jill      │ Fatigue  │ sigh*   │ 1     │
├───────────┼──────────┼─────────┼───────┤
│ john      │ Cough    │ achoo   │ 1     │
│ john      │ Fatigue  │ sigh*   │ 2     │
│ john      │ Fatigue  │ sleepy  │ 1     │
│ john      │ Headache │ pain*   │ 1     │
│ john      │ Headache │ ouch*   │ 1     │
╰───────────┴──────────┴─────────┴───────╯
  * This text has multiple associated labels.
""",
            stdout,
        )

    def test_ignored(self):
        """Verify that we show info on ignored notes"""
        stdout = self.run_cli("frequency", path=f"{self.DATA_DIR}/ignore")

        # Showing empty mentions felt like the most reasonable approach to the edge case of
        # "no text in the annotation" - but also disabling the term confusion warning for
        # empty mentions.
        self.assertEqual(
            """╭───────────┬───────┬─────────┬───────╮
│ Annotator │ Label │ Mention │ Count │
├───────────┼───────┼─────────┼───────┤
│ All       │ A     │         │ 2     │
│ All       │ B     │         │ 2     │
├───────────┼───────┼─────────┼───────┤
│ adam      │ A     │         │ 1     │
│ adam      │ B     │         │ 1     │
├───────────┼───────┼─────────┼───────┤
│ allison   │ A     │         │ 1     │
│ allison   │ B     │         │ 1     │
╰───────────┴───────┴─────────┴───────╯
  Ignoring 3 charts (3–5)
""",
            stdout,
        )

    def test_external(self):
        """Verify that we don't show external annotators"""
        stdout = self.run_cli("frequency", path=f"{self.DATA_DIR}/external")

        self.assertEqual(
            """╭───────────┬───────┬─────────┬───────╮
│ Annotator │ Label │ Mention │ Count │
├───────────┼───────┼─────────┼───────┤
│ All       │ happy │ woo     │ 1     │
│ All       │ sad   │ sigh    │ 1     │
├───────────┼───────┼─────────┼───────┤
│ human     │ happy │ woo     │ 1     │
│ human     │ sad   │ sigh    │ 1     │
╰───────────┴───────┴─────────┴───────╯
""",
            stdout,
        )

    def test_unused_labels(self):
        """Verify that we don't list mentions for labels that aren't in consideration"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "annotators": {"chris": 1},
                    "labels": ["Valid"],
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {"value": {"text": "good", "labels": ["Valid"]}},
                                    {"value": {"text": "bad", "labels": ["Invalid"]}},
                                ],
                            },
                        ],
                    },
                ],
            )
            stdout = self.run_cli("frequency", path=tmpdir)

            self.assertEqual(
                """╭───────────┬───────┬─────────┬───────╮
│ Annotator │ Label │ Mention │ Count │
├───────────┼───────┼─────────┼───────┤
│ All       │ Valid │ good    │ 1     │
├───────────┼───────┼─────────┼───────┤
│ chris     │ Valid │ good    │ 1     │
╰───────────┴───────┴─────────┴───────╯
""",
                stdout,
            )

    def test_spaces(self):
        """Verify that we handle text with surrounding spaces etc"""
        with tempfile.TemporaryDirectory() as tmpdir:
            common.write_json(
                f"{tmpdir}/config.json",
                {
                    "annotators": {"chris": 1},
                    "labels": ["LabelA"],
                },
            )
            common.write_json(
                f"{tmpdir}/labelstudio-export.json",
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {"value": {"text": "extra SPACES ", "labels": ["LabelA"]}},
                                    {"value": {"text": "\nextra spaces", "labels": ["LabelA"]}},
                                    {
                                        "value": {
                                            "text": "     Extra Spaces    ",
                                            "labels": ["LabelA"],
                                        }
                                    },
                                ],
                            },
                        ],
                    },
                ],
            )
            stdout = self.run_cli("frequency", path=tmpdir)

            self.assertEqual(
                """╭───────────┬────────┬──────────────┬───────╮
│ Annotator │ Label  │ Mention      │ Count │
├───────────┼────────┼──────────────┼───────┤
│ All       │ LabelA │ extra spaces │ 3     │
├───────────┼────────┼──────────────┼───────┤
│ chris     │ LabelA │ extra spaces │ 3     │
╰───────────┴────────┴──────────────┴───────╯
""",
                stdout,
            )

    def test_csv(self):
        """Verify that can print in CSV format"""
        stdout = self.run_cli("frequency", "--csv", path=f"{self.DATA_DIR}/external")

        self.assertEqual(
            [
                "annotator,label,mention,count",
                "All,happy,woo,1",
                "All,sad,sigh,1",
                "human,happy,woo,1",
                "human,sad,sigh,1",
            ],
            stdout.splitlines(),
        )

    def test_sublabels(self):
        output = self.run_cli("frequency", path=f"{self.DATA_DIR}/sublabels")
        self.assertEqual(
            output,
            """╭───────────┬────────────────────────────────┬───────────────────────┬───────╮
│ Annotator │ Label                          │ Mention               │ Count │
├───────────┼────────────────────────────────┼───────────────────────┼───────┤
│ All       │ Deceased                       │ death*                │ 1     │
│ All       │ Deceased → False               │ alive                 │ 3     │
│ All       │ Deceased → Datetime → 11/12/25 │ alive*                │ 2     │
│ All       │ Deceased → Datetime → 11/13/25 │ alive*                │ 1     │
│ All       │ Fungal → Confirmed             │ fungus found in lungs │ 3     │
│ All       │ Infection                      │ death*                │ 1     │
│ All       │ Infection → Confirmed          │ maybe got infected*   │ 1     │
│ All       │ Infection → Suspected          │ maybe got infected*   │ 2     │
├───────────┼────────────────────────────────┼───────────────────────┼───────┤
│ alice     │ Deceased                       │ death*                │ 1     │
│ alice     │ Deceased → False               │ alive                 │ 1     │
│ alice     │ Deceased → Datetime → 11/12/25 │ alive*                │ 1     │
│ alice     │ Fungal → Confirmed             │ fungus found in lungs │ 1     │
│ alice     │ Infection → Suspected          │ maybe got infected*   │ 1     │
├───────────┼────────────────────────────────┼───────────────────────┼───────┤
│ bob       │ Deceased → False               │ alive                 │ 1     │
│ bob       │ Deceased → Datetime → 11/13/25 │ alive*                │ 1     │
│ bob       │ Fungal → Confirmed             │ fungus found in lungs │ 1     │
│ bob       │ Infection                      │ death*                │ 1     │
│ bob       │ Infection → Suspected          │ maybe got infected*   │ 1     │
├───────────┼────────────────────────────────┼───────────────────────┼───────┤
│ carla     │ Deceased → False               │ alive                 │ 1     │
│ carla     │ Deceased → Datetime → 11/12/25 │ alive*                │ 1     │
│ carla     │ Fungal → Confirmed             │ fungus found in lungs │ 1     │
│ carla     │ Infection → Confirmed          │ maybe got infected*   │ 1     │
╰───────────┴────────────────────────────────┴───────────────────────┴───────╯
  * This text has multiple associated labels.
""",
        )



"""Tests for agree.py"""

import ddt

from chart_review import agree, defines
from tests import base


@ddt.ddt
class TestAgreement(base.TestCase):
    """Test case for basic agreement logic"""

    @ddt.data(
        (
            "alice",
            "bob",
            None,
            {
                "FN": [{1: "Cough"}],
                "FP": [{1: "Headache"}, {2: "Cough"}],
                "TN": [{1: "Fever"}, {2: "Headache"}],
                "TP": [{2: "Fever"}],
            },
        ),
        (
            "bob",
            "alice",
            {},
            {
                "FN": [{1: "Headache"}, {2: "Cough"}],
                "FP": [{1: "Cough"}],
                "TN": [{1: "Fever"}, {2: "Headache"}],
                "TP": [{2: "Fever"}],
            },
        ),
        (
            "alice",
            "bob",
            ["Cough"],
            {
                "FN": [{1: "Cough"}],
                "FP": [{2: "Cough"}],
                "TN": [],
                "TP": [],
            },
        ),
    )
    @ddt.unpack
    def test_confusion_matrix_counts(self, truth, annotator, labels, expected_matrix):
        """Verify that we can make a simple confusion matrix."""
        annotations = defines.ProjectAnnotations(
            labels={"Cough", "Fever", "Headache"},
            mentions={
                "alice": {1: {"Cough"}, 2: {"Fever"}},
                "bob": {1: {"Headache"}, 2: {"Cough", "Fever"}},
            },
        )
        notes = [1, 2]

        matrix = agree.confusion_matrix(annotations, truth, annotator, notes, labels=labels)
        self.assertEqual(expected_matrix, matrix)

    @ddt.data(
        # Examples pulled from https://en.wikipedia.org/wiki/Cohen's_kappa#Examples
        (
            {
                "FN": [{x: "Label"} for x in range(5)],
                "FP": [{x: "Label"} for x in range(10)],
                "TN": [{x: "Label"} for x in range(15)],
                "TP": [{x: "Label"} for x in range(20)],
            },
            0.4,
        ),
        (
            {
                "FN": [{x: "Label"} for x in range(15)],
                "FP": [{x: "Label"} for x in range(25)],
                "TN": [{x: "Label"} for x in range(15)],
                "TP": [{x: "Label"} for x in range(45)],
            },
            0.1304,
        ),
        (
            {
                "FN": [{x: "Label"} for x in range(35)],
                "FP": [{x: "Label"} for x in range(5)],
                "TN": [{x: "Label"} for x in range(35)],
                "TP": [{x: "Label"} for x in range(25)],
            },
            0.2593,
        ),
        # This example is from table 2 in https://pubmed.ncbi.nlm.nih.gov/12474424/
        (
            {
                "FN": [{x: "Label"} for x in range(6)],
                "FP": [{x: "Label"} for x in range(9)],
                "TN": [{x: "Label"} for x in range(26)],
                "TP": [{x: "Label"} for x in range(15)],
            },
            0.4444,
        ),
    )
    @ddt.unpack
    def test_kappa_score(self, matrix, expected_kappa):
        """Verify that we can score a matrix for kappa."""
        kappa = round(agree.score_kappa(matrix), 4)
        self.assertEqual(expected_kappa, kappa)



"""Methods for high-level accuracy calculations."""

import argparse
import os

import rich
import rich.box
import rich.table
import rich.text

from chart_review import agree, cli_utils, common, console_utils


def make_subparser(parser: argparse.ArgumentParser) -> None:
    cli_utils.add_project_args(parser)
    output_group = cli_utils.add_output_args(parser)
    output_group.add_argument("--save", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--verbose", action="store_true", help="show each chart’s labels")
    parser.add_argument("truth_annotator")
    parser.add_argument("annotator")
    parser.set_defaults(func=print_accuracy)


def print_accuracy(args: argparse.Namespace) -> None:
    """
    High-level accuracy calculation between two annotators.

    The results will be written to the project directory.
    """
    reader = cli_utils.get_cohort_reader(args)
    truth = args.truth_annotator
    annotator = args.annotator

    if truth not in reader.note_range:
        raise ValueError(f"Unrecognized annotator '{truth}'")
    if annotator not in reader.note_range:
        raise ValueError(f"Unrecognized annotator '{annotator}'")

    if truth == annotator:
        raise ValueError("Can’t compare the same annotator with themselves.")

    # Grab the intersection of ranges
    note_range = set(reader.note_range[truth])
    note_range &= set(reader.note_range[annotator])

    labels = sorted(reader.class_labels)

    # Calculate confusion matrices
    matrices = {None: reader.confusion_matrix(truth, annotator, note_range)}
    for label in labels:
        matrices[label] = reader.confusion_matrix(truth, annotator, note_range, label)

    # Now score them
    scores = {None: agree.score_matrix(matrices[None])}
    for label in labels:
        scores[label] = agree.score_matrix(matrices[label])

    console = rich.get_console()

    if args.verbose:
        # Print a table of each chart/label combo - useful for reviewing where an annotator
        # went wrong.
        table = cli_utils.create_table("Chart ID", "Label", "Classification")
        for note_id in sorted(note_range):
            table.add_section()
            for label in labels:
                for classification in ["TN", "TP", "FN", "FP"]:
                    if {note_id: label} in matrices[label][classification]:
                        style = "bold" if classification[0] == "F" else None  # highlight errors
                        class_text = rich.text.Text(classification, style=style)
                        table.add_row(str(note_id), str(label), class_text)
                        break
    else:
        # Normal F1/Kappa scores
        table = cli_utils.create_table(*agree.csv_header(), "Label", dense=True)
        table.add_row(*agree.csv_row_score(scores[None]), "*")
        for label in labels:
            table.add_row(*agree.csv_row_score(scores[label]), str(label))

    if args.csv:
        cli_utils.print_table_as_csv(table)
        return

    # OK we aren't printing a CSV file to stdout, so we can include a bit more explanation
    # as a little header to the real results.
    note_count = len(note_range)
    chart_word = "chart" if note_count == 1 else "charts"
    pretty_ranges = f" ({console_utils.pretty_note_range(note_range)})" if note_count > 0 else ""
    console.print(f"Comparing {note_count} {chart_word}{pretty_ranges}")
    console.print(f"Truth: {truth}")
    console.print(f"Annotator: {annotator}")

    if not args.verbose and labels:
        # Calculate Macro F1 as a convenience
        macro_f1 = sum(scores[label]["F1"] for label in labels) / len(labels)
        console.print(f"Macro F1: {round(macro_f1, 3)}")

    console.print()

    if args.save:  # deprecated/hidden since 2.0, but still supported for now
        output_stem = os.path.join(reader.project_dir, f"accuracy-{truth}-{annotator}")

        # Round before we hit disk
        for label_scores in scores.values():
            for key, value in label_scores.items():
                label_scores[key] = round(value, 3)

        # JSON: Historically, this has been formatted with the global label results intermixed
        # with the specific label names, so reproduce that historical formatting here.
        # Note: this could bite us if the user ever has a label like "Kappa", which is why the
        # above code avoids intermixing, but we'll keep this as-is for now.
        scores.update(scores[None])
        del scores[None]

        common.write_json(f"{output_stem}.json", {str(key): val for key, val in scores.items()})
        console.print(f"Wrote {output_stem}.json")

        # CSV: we should really use a .tsv suffix here, but keeping .csv for historical reasons
        common.write_text(f"{output_stem}.csv", agree.csv_table(scores, reader.class_labels))
        console.print(f"Wrote {output_stem}.csv")
    else:
        console.print(table)



import itertools
import os
import re
import sys
from collections.abc import Iterable

import yaml

from chart_review import defines, errors


class ProjectConfig:
    _NUMBER_REGEX = re.compile(r"\d+")
    _RANGE_REGEX = re.compile(r"\d+-\d+")

    def __init__(self, project_dir: str | None = None, config_path: str | None = None):
        """
        :param project_dir: str like /opt/labelstudio/study_name
        """
        self.project_dir = project_dir or "."
        try:
            self._data = self._load_config(config_path) or {}
        except Exception as exc:
            errors.exit_for_invalid_project(str(exc))

        # ** Annotators **
        # Internally, we're often dealing with numeric ID as the primary annotator identifier,
        # since that's what is stored in Label Studio. So that's what we return from this method.
        # But as humans writing config files, it's more natural to think of "name -> id".
        # So that's what we keep in the config, and we just reverse it here for convenience.
        self.annotators = defines.AnnotatorMap()
        self.external_annotations = {}
        for name, value in self._data.get("annotators", {}).items():
            if isinstance(value, int):  # real annotation layer in Label Studio
                self.annotators[value] = name
            else:  # fake/external annotation layer that we will inject
                self.external_annotations[name] = value

        # ** Note ranges **
        # Handle some extra syntax like 1-3 == [1, 2, 3]
        self.note_ranges: dict[str, list[int]] = self._data.get("ranges", {})
        for key, values in self.note_ranges.items():
            self.note_ranges[key] = list(self._parse_note_range(values))

        # ** Implied labels **
        self.implied_labels = defines.ImpliedLabels()
        for key, value in self._data.get("implied-labels", {}).items():
            # Coerce single labels into a set
            if not isinstance(value, list):
                value = {value}
            source_matcher = defines.LabelMatcher(key)
            self.implied_labels[source_matcher] = set(defines.Label.parse(x) for x in value)

        # ** Grouped labels **
        self.grouped_labels = defines.GroupedLabels()
        for key, value in self._data.get("grouped-labels", {}).items():
            # Coerce single labels into a set
            if not isinstance(value, list):
                value = {value}
            group_label = defines.Label.parse(key)
            self.grouped_labels[group_label] = defines.LabelMatcher(*value)

    def path(self, filename: str) -> str:
        return os.path.join(self.project_dir, filename)

    @staticmethod
    def _read_yaml(path) -> dict:
        with open(path, encoding="utf8") as f:
            return yaml.safe_load(f)

    def _load_config(self, config_path: str | None) -> dict:
        if config_path is None:
            # Support config.json in case folks prefer that
            try:
                config = self._read_yaml(self.path("config.json"))
            except FileNotFoundError:
                try:
                    config = self._read_yaml(self.path("config.yaml"))
                except FileNotFoundError:
                    config = {}
        else:
            # Don't resolve config_path relative to the project dir, because
            # this will have come from the command line and will resolve relative to `pwd`.
            config = self._read_yaml(config_path)

        # Do some minimal validation on the config file
        if not isinstance(config, dict):
            raise ValueError("Config file is not in the expected dictionary format.")

        return config

    def _parse_note_range(self, value: str | int | list[str | int]) -> Iterable[int]:
        if isinstance(value, list):
            return list(itertools.chain.from_iterable(self._parse_note_range(v) for v in value))
        elif isinstance(value, int):
            return [value]
        elif self._NUMBER_REGEX.fullmatch(value):
            return [int(value)]
        elif self._RANGE_REGEX.fullmatch(value):
            edges = value.split("-")
            return range(int(edges[0]), int(edges[1]) + 1)
        elif value in self.note_ranges:
            return self._parse_note_range(
                self.note_ranges[value]
            )  # warning: no guards against infinite recursion
        else:
            print(f"Unknown note range '{value}'", file=sys.stderr)
            return []

    @property
    def class_labels(self) -> defines.LabelSet:
        return set(defines.Label.parse(x) for x in self._data.setdefault("labels", []))

    @property
    def ignore(self) -> set[str]:
        return set(self._data.setdefault("ignore", []))



"""Tests for simplify.py"""

import ddt

from chart_review import defines, simplify
from tests import base


@ddt.ddt
class TestSimplify(base.TestCase):
    """Test case for annotation simplification"""

    @ddt.data(
        ({"Cat"}, {"Animal", "Cat", "Pet"}),  # Basic expansion
        ({"Animal", "Dangerous", "Venomous"}, {"Animal", "Risky"}),  # Basic grouping
        ({"Lion"}, {"Animal", "Cat", "Lion", "Pet", "Risky"}),  # Two layers deep, plus grouping
    )
    @ddt.unpack
    def test_simplify_mentions(self, input_labels, expected_labels):
        """Verify that we handle implied and grouped labels."""
        annotations = defines.ProjectAnnotations(
            # Note that the group names are not part of this label set yet.
            # This set generally starts with just the human-tagged input labels.
            labels=base.labels({"Animal", "Cat", "Dangerous", "Lion", "Pet", "Venomous"}),
            mentions={
                "katherine": {1: base.labels(input_labels)},
            },
        )
        simplify.simplify_mentions(
            annotations,
            implied_labels={
                base.LabelMatcher("Cat"): base.labels({"Animal", "Pet"}),
                base.LabelMatcher("Lion"): base.labels({"Cat", "Dangerous"}),
            },
            grouped_labels={
                # Dangerous overlaps with implied labels above, to test ordering
                base.Label("Risky"): base.LabelMatcher("Dangerous", "Venomous"),
            },
        )
        self.assertEqual(base.labels(expected_labels), annotations.mentions["katherine"][1])
        self.assertEqual(base.labels({"Animal", "Cat", "Lion", "Pet", "Risky"}), annotations.labels)

    def test_binary_grouping(self):
        """Verify that grouping all labels down to a simple yes/no works as expected."""
        annotations = defines.ProjectAnnotations(
            labels=base.labels({"Blue", "Green", "Red"}),
            mentions={
                "paint-bot": {
                    1: base.labels({"Blue"}),
                    2: set(),
                    3: base.labels({"Red", "Green"}),
                },
            },
        )
        simplify.simplify_mentions(
            annotations,
            implied_labels={},
            grouped_labels={base.Label("Painted"): base.LabelMatcher("Blue", "Green", "Red")},
        )
        self.assertEqual(
            {1: base.labels({"Painted"}), 2: set(), 3: base.labels({"Painted"})},
            annotations.mentions["paint-bot"],
        )
        self.assertEqual(base.labels({"Painted"}), annotations.labels)

    @ddt.data(
        (["A|B|C"], ["A|B|C"], True),
        (["A|B|D"], ["A|B|C"], False),
        (["A|B|C"], ["A|B"], True),
        (["A|C|D"], ["A|B"], False),
        (["A|B|C"], ["A"], True),
        (["B|C|D"], ["A"], False),
    )
    @ddt.unpack
    def test_sublabel_implied_matching(self, labels, config, add_to_expected):
        labels = base.labels(labels)
        annotations = defines.ProjectAnnotations(labels=labels, mentions={"alice": {1: labels}})
        simplify.simplify_mentions(
            annotations,
            implied_labels={base.LabelMatcher(*config): base.labels({"Implied"})},
            grouped_labels={},
        )
        if add_to_expected:
            labels.add(base.Label("Implied"))
        self.assertEqual(annotations.mentions["alice"][1], labels)
        self.assertEqual(annotations.labels, labels)

    @ddt.data(
        (["A"], ["Group", "H"]),
        (["A|B"], ["Group", "H", "A|M|N"]),
        (["A|B|C"], ["Group", "H", "A|B|D", "A|M|N"]),
    )
    @ddt.unpack
    def test_sublabel_group_matching(self, config, expected):
        all_labels = base.labels({"A|B|C", "A|B|D", "A|M|N", "H"})
        annotations = defines.ProjectAnnotations(
            labels=all_labels,
            mentions={"alice": {1: all_labels}},
        )
        simplify.simplify_mentions(
            annotations,
            implied_labels={},
            grouped_labels={base.Label("Group"): base.LabelMatcher(*config)},
        )
        self.assertEqual(annotations.mentions["alice"][1], base.labels(expected))
        self.assertEqual(annotations.labels, base.labels(expected))



import argparse

import rich
import rich.box
import rich.table
import rich.text

from chart_review import cli_utils, console_utils


def make_subparser(parser: argparse.ArgumentParser) -> None:
    cli_utils.add_project_args(parser)
    cli_utils.add_output_args(parser)
    parser.set_defaults(func=print_frequency)


def print_frequency(args: argparse.Namespace) -> None:
    """
    Print counts of each text mention.
    """
    reader = cli_utils.get_cohort_reader(args)

    frequencies = {}  # annotator -> label -> text -> count
    all_annotator_frequencies = {}  # label -> text -> count
    text_labels = {}  # text -> label namespace -> labelset (to flag term confusion)
    for annotator in reader.annotations.original_text_mentions:
        annotator_mentions = reader.annotations.original_text_mentions[annotator]
        for labeled_texts in annotator_mentions.values():
            for labeled_text in labeled_texts:
                text = (labeled_text.text or "").strip().casefold()
                for label in labeled_text.labels:
                    if label in reader.annotations.labels:
                        # Count the mention for this annotator
                        label_to_text = frequencies.setdefault(annotator, {})
                        text_to_count = label_to_text.setdefault(label, {})
                        text_to_count[text] = text_to_count.get(text, 0) + 1

                        # Count the mention for our running all-annotators total
                        all_text_to_count = all_annotator_frequencies.setdefault(label, {})
                        all_text_to_count[text] = all_text_to_count.get(text, 0) + 1

                        # And finally, add it to our running term-confusion tracker
                        sublabel_to_labels = text_labels.setdefault(text, {})
                        sublabel_to_labels.setdefault(label.namespace(), set()).add(label)

    # Now group up the data into a formatted table
    table = cli_utils.create_table("Annotator", "Label", "Mention", "Count")
    has_term_confusion = False  # whether multiple labels are used for the same text

    # Helper method to add all the info for a single annotator to our table
    def add_annotator_to_table(name, label_to_text: dict) -> None:
        nonlocal has_term_confusion
        table.add_section()
        for label in sorted(label_to_text):
            text_to_count = label_to_text[label]
            for text, count in sorted(
                text_to_count.items(), key=lambda t: (t[1], t[0]), reverse=True
            ):
                if not args.csv and text:
                    if len(text_labels[text][label.namespace()]) > 1:
                        text = rich.text.Text(text + "*", style="bold")
                        has_term_confusion = True
                table.add_row(name, str(label), text, f"{count:,}")

    # Add each annotator
    add_annotator_to_table(rich.text.Text("All", style="italic"), all_annotator_frequencies)
    for annotator in sorted(frequencies, key=str.casefold):
        add_annotator_to_table(annotator, frequencies[annotator])

    if args.csv:
        cli_utils.print_table_as_csv(table)
    else:
        rich.get_console().print(table)
        console_utils.print_ignored_charts(reader)
        if has_term_confusion:
            rich.get_console().print(
                "  * This text has multiple associated labels.", style="italic"
            )



"""Helper methods for CLI parsing."""

import argparse
import csv
import sys

import rich.box
import rich.table

from chart_review import cohort, config


def add_project_args(parser: argparse.ArgumentParser, is_global: bool = False) -> None:
    group = parser.add_argument_group("configuration")
    group.add_argument(
        "--project-dir",
        "-p",
        default=None if is_global else argparse.SUPPRESS,
        metavar="DIR",
        help=(
            "directory holding project files, like labelstudio-export.json (default: current dir)"
        ),
    )
    group.add_argument(
        "--config",
        "-c",
        default=None if is_global else argparse.SUPPRESS,
        metavar="PATH",
        help="config file (default: [project-dir]/config.yaml)",
    )


def add_output_args(parser: argparse.ArgumentParser):
    """Returns an exclusive option group if you want to add custom output arguments"""
    group = parser.add_argument_group("output")
    exclusive = group.add_mutually_exclusive_group()
    exclusive.add_argument("--csv", action="store_true", help="print results in CSV format")
    return exclusive


def get_cohort_reader(args: argparse.Namespace) -> cohort.CohortReader:
    proj_config = config.ProjectConfig(project_dir=args.project_dir, config_path=args.config)
    return cohort.CohortReader(proj_config)


def create_table(*headers, dense: bool = False) -> rich.table.Table:
    """
    Creates a table with standard chart-review formatting.

    You can use your own table formatting if you have particular needs,
    but this should be your default table creator.
    """
    if dense:
        table = rich.table.Table(*headers, box=None, pad_edge=False)
    else:
        table = rich.table.Table(box=rich.box.ROUNDED)
        for header in headers:
            table.add_column(header, overflow="fold")
    return table


def print_table_as_csv(table: rich.table.Table) -> None:
    """Prints a Rich table as a CSV to stdout"""
    writer = csv.writer(sys.stdout)

    # First the headers
    headers = [str(col.header).lower().replace(" ", "_") for col in table.columns]
    writer.writerow(headers)

    # And then each row
    cells_by_row = zip(*[col.cells for col in table.columns])
    for row in cells_by_row:
        writer.writerow(row)



import argparse

import rich
import rich.box
import rich.table
import rich.text

from chart_review import cli_utils, console_utils, defines


def make_subparser(parser: argparse.ArgumentParser) -> None:
    cli_utils.add_project_args(parser)
    cli_utils.add_output_args(parser)
    parser.set_defaults(func=print_labels)


def print_labels(args: argparse.Namespace) -> None:
    """Show label information on the console."""
    reader = cli_utils.get_cohort_reader(args)

    # Calculate all label counts for each annotator
    label_names = sorted(reader.class_labels)
    label_notes: dict[str, dict[str, defines.NoteSet]] = {}  # annotator -> label -> note IDs
    any_annotator_note_sets: dict[str, defines.NoteSet] = {}
    for annotator, mentions in reader.annotations.mentions.items():
        label_notes[annotator] = {}
        for name in label_names:
            note_ids = {note_id for note_id, labels in mentions.items() if name in labels}
            label_notes[annotator][name] = note_ids
            any_annotator_note_sets.setdefault(name, defines.NoteSet()).update(note_ids)

    label_table = cli_utils.create_table("Annotator", "Label", "Chart Count")

    # First add summary entries, for counts across the union of all annotators
    for name in label_names:
        count = f"{len(any_annotator_note_sets.get(name, {})):,}"
        label_table.add_row(rich.text.Text("Any", style="italic"), str(name), count)

    # Now do each annotator as their own little boxed section
    for annotator in sorted(label_notes.keys(), key=str.casefold):
        label_table.add_section()
        for name, note_set in label_notes[annotator].items():
            count = str(len(note_set))
            label_table.add_row(annotator, str(name), count)

    if args.csv:
        cli_utils.print_table_as_csv(label_table)
    else:
        rich.get_console().print(label_table)
        console_utils.print_ignored_charts(reader)
